# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Unit tests for the `iris.fileformats.pp_load_rules._all_other_rules` function.

"""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

from unittest import mock

import numpy as np
from cf_units import Unit, CALENDAR_360_DAY
from cftime import datetime as nc_datetime

from iris.fileformats.pp_load_rules import _all_other_rules
from iris.fileformats.pp import SplittableInt
from iris.coords import CellMethod, DimCoord, AuxCoord
from iris.tests.unit.fileformats import TestField


# iris.fileformats.pp_load_rules._all_other_rules() returns a tuple of
# of various metadata. This constant is the index into this
# tuple to obtain the cell methods.
CELL_METHODS_INDEX = 5
DIM_COORDS_INDEX = 6
AUX_COORDS_INDEX = 7


class TestCellMethods(tests.IrisTest):
    def test_time_mean(self):
        # lbproc = 128 -> mean
        # lbtim.ib = 2 -> simple t1 to t2 interval.
        field = mock.MagicMock(lbproc=128, lbtim=mock.Mock(ia=0, ib=2, ic=3))
        res = _all_other_rules(field)[CELL_METHODS_INDEX]
        expected = [CellMethod("mean", "time")]
        self.assertEqual(res, expected)

    def test_hourly_mean(self):
        # lbtim.ia = 1 -> hourly
        field = mock.MagicMock(lbproc=128, lbtim=mock.Mock(ia=1, ib=2, ic=3))
        res = _all_other_rules(field)[CELL_METHODS_INDEX]
        expected = [CellMethod("mean", "time", "1 hour")]
        self.assertEqual(res, expected)

    def test_daily_mean(self):
        # lbtim.ia = 24 -> daily
        field = mock.MagicMock(lbproc=128, lbtim=mock.Mock(ia=24, ib=2, ic=3))
        res = _all_other_rules(field)[CELL_METHODS_INDEX]
        expected = [CellMethod("mean", "time", "24 hour")]
        self.assertEqual(res, expected)

    def test_custom_max(self):
        field = mock.MagicMock(lbproc=8192, lbtim=mock.Mock(ia=47, ib=2, ic=3))
        res = _all_other_rules(field)[CELL_METHODS_INDEX]
        expected = [CellMethod("maximum", "time", "47 hour")]
        self.assertEqual(res, expected)

    def test_daily_min(self):
        # lbproc = 4096 -> min
        field = mock.MagicMock(lbproc=4096, lbtim=mock.Mock(ia=24, ib=2, ic=3))
        res = _all_other_rules(field)[CELL_METHODS_INDEX]
        expected = [CellMethod("minimum", "time", "24 hour")]
        self.assertEqual(res, expected)

    def test_time_mean_over_multiple_years(self):
        # lbtim.ib = 3 -> interval within a year, over multiple years.
        field = mock.MagicMock(lbproc=128, lbtim=mock.Mock(ia=0, ib=3, ic=3))
        res = _all_other_rules(field)[CELL_METHODS_INDEX]
        expected = [
            CellMethod("mean within years", "time"),
            CellMethod("mean over years", "time"),
        ]
        self.assertEqual(res, expected)

    def test_hourly_mean_over_multiple_years(self):
        field = mock.MagicMock(lbproc=128, lbtim=mock.Mock(ia=1, ib=3, ic=3))
        res = _all_other_rules(field)[CELL_METHODS_INDEX]
        expected = [
            CellMethod("mean within years", "time", "1 hour"),
            CellMethod("mean over years", "time"),
        ]
        self.assertEqual(res, expected)

    def test_climatology_max(self):
        field = mock.MagicMock(lbproc=8192, lbtim=mock.Mock(ia=24, ib=3, ic=3))
        res = _all_other_rules(field)[CELL_METHODS_INDEX]
        expected = [CellMethod("maximum", "time")]
        self.assertEqual(res, expected)

    def test_climatology_min(self):
        field = mock.MagicMock(lbproc=4096, lbtim=mock.Mock(ia=24, ib=3, ic=3))
        res = _all_other_rules(field)[CELL_METHODS_INDEX]
        expected = [CellMethod("minimum", "time")]
        self.assertEqual(res, expected)

    def test_other_lbtim_ib(self):
        # lbtim.ib = 5 -> non-specific aggregation
        field = mock.MagicMock(lbproc=4096, lbtim=mock.Mock(ia=24, ib=5, ic=3))
        res = _all_other_rules(field)[CELL_METHODS_INDEX]
        expected = [CellMethod("minimum", "time")]
        self.assertEqual(res, expected)

    def test_multiple_unordered_lbprocs(self):
        field = mock.MagicMock(
            lbproc=192,
            bzx=0,
            bdx=1,
            lbnpt=3,
            lbrow=3,
            lbtim=mock.Mock(ia=24, ib=5, ic=3),
            lbcode=SplittableInt(1),
            x_bounds=None,
            _x_coord_name=lambda: "longitude",
            _y_coord_name=lambda: "latitude",
        )
        res = _all_other_rules(field)[CELL_METHODS_INDEX]
        expected = [
            CellMethod("mean", "time"),
            CellMethod("mean", "longitude"),
        ]
        self.assertEqual(res, expected)

    def test_multiple_unordered_rotated_lbprocs(self):
        field = mock.MagicMock(
            lbproc=192,
            bzx=0,
            bdx=1,
            lbnpt=3,
            lbrow=3,
            lbtim=mock.Mock(ia=24, ib=5, ic=3),
            lbcode=SplittableInt(101),
            x_bounds=None,
            _x_coord_name=lambda: "grid_longitude",
            _y_coord_name=lambda: "grid_latitude",
        )
        res = _all_other_rules(field)[CELL_METHODS_INDEX]
        expected = [
            CellMethod("mean", "time"),
            CellMethod("mean", "grid_longitude"),
        ]
        self.assertEqual(res, expected)


class TestCrossSectionalTime(TestField):
    def test_lbcode3x23(self):
        time_bounds = np.array(
            [[0.875, 1.125], [1.125, 1.375], [1.375, 1.625], [1.625, 1.875]]
        )
        field = mock.MagicMock(
            lbproc=0,
            bzx=0,
            bdx=0,
            lbnpt=3,
            lbrow=4,
            t1=nc_datetime(2000, 1, 2, hour=0, minute=0, second=0),
            t2=nc_datetime(2000, 1, 3, hour=0, minute=0, second=0),
            lbtim=mock.Mock(ia=1, ib=2, ic=2),
            lbcode=SplittableInt(
                31323, {"iy": slice(0, 2), "ix": slice(2, 4)}
            ),
            x_bounds=None,
            y_bounds=time_bounds,
            _x_coord_name=lambda: "longitude",
            _y_coord_name=lambda: "latitude",
        )

        spec = [
            "lbtim",
            "lbcode",
            "lbrow",
            "lbnpt",
            "lbproc",
            "lbsrce",
            "lbuser",
            "bzx",
            "bdx",
            "bdy",
            "bmdi",
            "t1",
            "t2",
            "stash",
            "x_bounds",
            "y_bounds",
            "_x_coord_name",
            "_y_coord_name",
        ]
        field.mock_add_spec(spec)
        res = _all_other_rules(field)[DIM_COORDS_INDEX]

        expected_time_points = np.array([1, 1.25, 1.5, 1.75]) + (2000 * 360)
        expected_unit = Unit(
            "days since 0000-01-01 00:00:00", calendar=CALENDAR_360_DAY
        )
        expected = [
            (
                DimCoord(
                    expected_time_points,
                    standard_name="time",
                    units=expected_unit,
                    bounds=time_bounds,
                ),
                0,
            )
        ]
        self.assertCoordsAndDimsListsMatch(res, expected)


class TestLBTIMx2x_ZeroYears(TestField):

    _spec = [
        "lbtim",
        "lbcode",
        "lbrow",
        "lbnpt",
        "lbproc",
        "lbsrce",
        "lbhem",
        "lbuser",
        "bzx",
        "bdx",
        "bdy",
        "bmdi",
        "t1",
        "t2",
        "stash",
        "x_bounds",
        "y_bounds",
        "_x_coord_name",
        "_y_coord_name",
    ]

    def _make_field(
        self,
        lbyr=0,
        lbyrd=0,
        lbmon=3,
        lbmond=3,
        lbft=0,
        bdx=1,
        bdy=1,
        bmdi=0,
        ia=0,
        ib=2,
        ic=1,
        lbcode=SplittableInt(3),
    ):
        return mock.MagicMock(
            lbyr=lbyr,
            lbyrd=lbyrd,
            lbmon=lbmon,
            lbmond=lbmond,
            lbft=lbft,
            bdx=bdx,
            bdy=bdy,
            bmdi=bmdi,
            lbtim=mock.Mock(ia=ia, ib=ib, ic=ic),
            lbcode=lbcode,
        )

    def test_month_coord(self):
        field = self._make_field()
        field.mock_add_spec(self._spec)
        res = _all_other_rules(field)[AUX_COORDS_INDEX]

        expected = [
            (AuxCoord(3, long_name="month_number", units="1"), None),
            (AuxCoord("Mar", long_name="month", units=Unit("no unit")), None),
            (
                DimCoord(
                    points=0,
                    standard_name="forecast_period",
                    units=Unit("hours"),
                ),
                None,
            ),
        ]
        self.assertCoordsAndDimsListsMatch(res, expected)

    def test_diff_month(self):
        field = self._make_field(lbmon=3, lbmond=4)
        field.mock_add_spec(self._spec)
        res = _all_other_rules(field)[AUX_COORDS_INDEX]

        self.assertCoordsAndDimsListsMatch(res, [])

    def test_nonzero_year(self):
        field = self._make_field(lbyr=1)
        field.mock_add_spec(self._spec)
        res = _all_other_rules(field)[AUX_COORDS_INDEX]

        self.assertCoordsAndDimsListsMatch(res, [])

    def test_nonzero_yeard(self):
        field = self._make_field(lbyrd=1)
        field.mock_add_spec(self._spec)
        res = _all_other_rules(field)[AUX_COORDS_INDEX]

        self.assertCoordsAndDimsListsMatch(res, [])


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Management of common state and behaviour for cube and coordinate data.

"""

import copy
import numpy as np
import numpy.ma as ma

from iris._lazy_data import as_concrete_data, as_lazy_data, is_lazy_data
from iris.util import array_equal


class DataManager:
    """
    Provides a well defined API for management of real or lazy data.

    """

    def __init__(self, data):
        """
        Create a data manager for the specified data.

        Args:

        * data:
            The :class:`~numpy.ndarray` or :class:`~numpy.ma.core.MaskedArray`
            real data, or :class:`~dask.array.core.Array` lazy data to be
            managed.

        """
        # Initialise the instance.
        self._lazy_array = None
        self._real_array = None

        # Assign the data payload to be managed.
        self.data = data

        # Enforce the manager contract.
        self._assert_axioms()

    def __copy__(self):
        """
        Forbid :class:`~iris._data_manager.DataManager` instance
        shallow-copy support.

        """
        name = type(self).__name__
        emsg = (
            "Shallow-copy of {!r} is not permitted. Use "
            "copy.deepcopy() or {}.copy() instead."
        )
        raise copy.Error(emsg.format(name, name))

    def __deepcopy__(self, memo):
        """
        Allow :class:`~iris._data_manager.DataManager` instance
        deepcopy support.

        Args:

        * memo:
            :class:`copy` memo dictionary.

        """
        return self._deepcopy(memo)

    def __eq__(self, other):
        """
        Perform :class:`~iris._data_manager.DataManager` instance equality.
        Note that, this is explicitly not a lazy operation and will load any
        lazy payload to determine the equality result.

        Comparison is strict with regards to lazy or real managed payload,
        the realised_dtype, the dtype of the payload, the fill-value and the
        payload content.

        Args:

        * other:
            The :class:`~iris._data_manager.DataManager` instance to
            compare with.

        Returns:
            Boolean.

        """
        result = NotImplemented

        if isinstance(other, type(self)):
            result = False
            same_lazy = self.has_lazy_data() == other.has_lazy_data()
            same_dtype = self.dtype == other.dtype
            if same_lazy and same_dtype:
                result = array_equal(self.core_data(), other.core_data())

        return result

    def __ne__(self, other):
        """
        Perform :class:`~iris._data_manager.DataManager` instance inequality.
        Note that, this is explicitly not a lazy operation and will load any
        lazy payload to determine the inequality result.

        Args:

        * other:
            The :class:`~iris._data_manager.DataManager` instance to
            compare with.

        Returns:
            Boolean.

        """
        result = self.__eq__(other)

        if result is not NotImplemented:
            result = not result

        return result

    def __repr__(self):
        """
        Returns an string representation of the instance.

        """
        fmt = "{cls}({data!r})"
        result = fmt.format(data=self.core_data(), cls=type(self).__name__)

        return result

    def _assert_axioms(self):
        """
        Definition of the manager state, that should never be violated.

        """
        # Ensure there is a valid data state.
        is_lazy = self._lazy_array is not None
        is_real = self._real_array is not None
        emsg = "Unexpected data state, got {}lazy and {}real data."
        state = is_lazy ^ is_real
        assert state, emsg.format(
            "" if is_lazy else "no ", "" if is_real else "no "
        )

    def _deepcopy(self, memo, data=None):
        """
        Perform a deepcopy of the :class:`~iris._data_manager.DataManager`
        instance.

        Args:

        * memo:
            :class:`copy` memo dictionary.

        Kwargs:

        * data:
            Replacement data to substitute the currently managed
            data with.

        Returns:
            :class:`~iris._data_manager.DataManager` instance.

        """
        try:
            if data is None:
                # Copy the managed data.
                if self.has_lazy_data():
                    data = copy.deepcopy(self._lazy_array, memo)
                else:
                    data = self._real_array.copy()
            else:
                # Check that the replacement data is valid relative to
                # the currently managed data.
                dm_check = DataManager(self.core_data())
                dm_check.data = data
                # If the replacement data is valid, then use it but
                # without copying it.
            result = DataManager(data)
        except ValueError as error:
            emsg = "Cannot copy {!r} - {}"
            raise ValueError(emsg.format(type(self).__name__, error))

        return result

    @property
    def data(self):
        """
        Returns the real data. Any lazy data being managed will be realised.

        Returns:
            :class:`~numpy.ndarray` or :class:`numpy.ma.core.MaskedArray`.

        """
        if self.has_lazy_data():
            try:
                # Realise the lazy data.
                result = as_concrete_data(self._lazy_array)
                # Assign the realised result.
                self._real_array = result
                # Reset the lazy data and the realised dtype.
                self._lazy_array = None
            except MemoryError:
                emsg = (
                    "Failed to realise the lazy data as there was not "
                    "enough memory available.\n"
                    "The data shape would have been {!r} with {!r}.\n "
                    "Consider freeing up variables or indexing the data "
                    "before trying again."
                )
                raise MemoryError(emsg.format(self.shape, self.dtype))

        # Check the manager contract, as the managed data has changed.
        self._assert_axioms()

        return self._real_array

    @data.setter
    def data(self, data):
        """
        Replaces the currently managed data with the specified data, which must
        be of an equivalent shape.

        Note that, the only shape promotion permitted is for 0-dimensional
        scalar data to be replaced with a single item 1-dimensional data.

        Args:

        * data:
            The :class:`~numpy.ndarray` or :class:`~numpy.ma.core.MaskedArray`
            real data, or :class:`~dask.array.core.Array` lazy data to be
            managed.

        """
        # Ensure we have numpy-like data.
        if not (hasattr(data, "shape") and hasattr(data, "dtype")):
            data = np.asanyarray(data)

        # Determine whether the class instance has been created,
        # as this method is called from within the __init__.
        init_done = (
            self._lazy_array is not None or self._real_array is not None
        )

        if init_done and self.shape != data.shape:
            # The _ONLY_ data reshape permitted is converting a 0-dimensional
            # array i.e. self.shape == () into a 1-dimensional array of length
            # one i.e. data.shape == (1,)
            if self.shape or data.shape != (1,):
                emsg = "Require data with shape {!r}, got {!r}."
                raise ValueError(emsg.format(self.shape, data.shape))

        # Set lazy or real data, and reset the other.
        if is_lazy_data(data):
            self._lazy_array = data
            self._real_array = None
        else:
            if not ma.isMaskedArray(data):
                # Coerce input data to ndarray (including ndarray subclasses).
                data = np.asarray(data)
            if isinstance(data, ma.core.MaskedConstant):
                # Promote to a masked array so that the fill-value is
                # writeable to the data owner.
                data = ma.array(data.data, mask=data.mask, dtype=data.dtype)
            self._lazy_array = None
            self._real_array = data

        # Check the manager contract, as the managed data has changed.
        self._assert_axioms()

    @property
    def dtype(self):
        """
        The dtype of the realised lazy data or the dtype of the real data.

        """
        return self.core_data().dtype

    @property
    def ndim(self):
        """
        The number of dimensions covered by the data being managed.

        """
        return self.core_data().ndim

    @property
    def shape(self):
        """
        The shape of the data being managed.

        """
        return self.core_data().shape

    def copy(self, data=None):
        """
        Returns a deep copy of this :class:`~iris._data_manager.DataManager`
        instance.

        Kwargs:

        * data:
            Replace the data of the copy with this data.

        Returns:
            A copy :class:`~iris._data_manager.DataManager` instance.

        """
        memo = {}
        return self._deepcopy(memo, data=data)

    def core_data(self):
        """
        If real data is being managed, then return the :class:`~numpy.ndarray`
        or :class:`numpy.ma.core.MaskedArray`. Otherwise, return the lazy
        :class:`~dask.array.core.Array`.

        Returns:
            The real or lazy data.

        """
        if self.has_lazy_data():
            result = self._lazy_array
        else:
            result = self._real_array

        return result

    def has_lazy_data(self):
        """
        Determine whether lazy data is being managed.

        Returns:
            Boolean.

        """
        return self._lazy_array is not None

    def lazy_data(self):
        """
        Return the lazy representation of the managed data.

        If only real data is being managed, then return a lazy
        representation of that real data.

        Returns:
            :class:`~dask.array.core.Array`

        .. note::
            This method will never realise any lazy data.

        """
        if self.has_lazy_data():
            result = self._lazy_array
        else:
            result = as_lazy_data(self._real_array)

        return result



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Provides testing capabilities and customisations specific to Iris.

.. note:: This module needs to control the matplotlib backend, so it
          **must** be imported before ``matplotlib.pyplot``.

The primary class for this module is :class:`IrisTest`.

By default, this module sets the matplotlib backend to "agg". But when
this module is imported it checks ``sys.argv`` for the flag "-d". If
found, it is removed from ``sys.argv`` and the matplotlib backend is
switched to "tkagg" to allow the interactive visual inspection of
graphical test results.

"""

import codecs
import collections
from collections.abc import Mapping
import contextlib
import datetime
import difflib
import filecmp
import functools
import gzip
import inspect
import json
import io
import math
import os
import os.path
import re
import shutil
import subprocess
import sys
import unittest
from unittest import mock
import threading
import warnings
import xml.dom.minidom
import zlib

import filelock
import numpy as np
import numpy.ma as ma
import requests

import iris.cube
import iris.config
import iris.util

# Test for availability of matplotlib.
# (And remove matplotlib as an iris.tests dependency.)
try:
    import matplotlib

    # Override any user settings e.g. from matplotlibrc file.
    matplotlib.rcdefaults()
    # Set backend *after* rcdefaults, as we don't want that overridden (#3846).
    matplotlib.use("agg")
    # Standardise the figure size across matplotlib versions.
    # This permits matplotlib png image comparison.
    matplotlib.rcParams["figure.figsize"] = [8.0, 6.0]
    import matplotlib.pyplot as plt
except ImportError:
    MPL_AVAILABLE = False
else:
    MPL_AVAILABLE = True

try:
    from osgeo import gdal  # noqa
except ImportError:
    GDAL_AVAILABLE = False
else:
    GDAL_AVAILABLE = True

try:
    import iris_sample_data  # noqa
except ImportError:
    SAMPLE_DATA_AVAILABLE = False
else:
    SAMPLE_DATA_AVAILABLE = True

try:
    import nc_time_axis  # noqa

    NC_TIME_AXIS_AVAILABLE = True
except ImportError:
    NC_TIME_AXIS_AVAILABLE = False

try:
    # Added a timeout to stop the call to requests.get hanging when running
    # on a platform which has restricted/no internet access.
    requests.get("https://github.com/SciTools/iris", timeout=10.0)
    INET_AVAILABLE = True
except requests.exceptions.ConnectionError:
    INET_AVAILABLE = False

try:
    import stratify  # noqa

    STRATIFY_AVAILABLE = True
except ImportError:
    STRATIFY_AVAILABLE = False

#: Basepath for test results.
_RESULT_PATH = os.path.join(os.path.dirname(__file__), "results")
#: Default perceptual hash size.
_HASH_SIZE = 16
#: Default maximum perceptual hash hamming distance.
_HAMMING_DISTANCE = 2

if "--data-files-used" in sys.argv:
    sys.argv.remove("--data-files-used")
    fname = "/var/tmp/all_iris_test_resource_paths.txt"
    print("saving list of files used by tests to %s" % fname)
    _EXPORT_DATAPATHS_FILE = open(fname, "w")
else:
    _EXPORT_DATAPATHS_FILE = None


if "--create-missing" in sys.argv:
    sys.argv.remove("--create-missing")
    print("Allowing creation of missing test results.")
    os.environ["IRIS_TEST_CREATE_MISSING"] = "true"


# Whether to display matplotlib output to the screen.
_DISPLAY_FIGURES = False

if MPL_AVAILABLE and "-d" in sys.argv:
    sys.argv.remove("-d")
    plt.switch_backend("tkagg")
    _DISPLAY_FIGURES = True

# Threading non re-entrant blocking lock to ensure thread-safe plotting.
_lock = threading.Lock()


def main():
    """A wrapper for unittest.main() which adds iris.test specific options to the help (-h) output."""
    if "-h" in sys.argv or "--help" in sys.argv:
        stdout = sys.stdout
        buff = io.StringIO()
        # NB. unittest.main() raises an exception after it's shown the help text
        try:
            sys.stdout = buff
            unittest.main()
        finally:
            sys.stdout = stdout
            lines = buff.getvalue().split("\n")
            lines.insert(9, "Iris-specific options:")
            lines.insert(
                10,
                "  -d                   Display matplotlib figures (uses tkagg).",
            )
            lines.insert(
                11,
                "                       NOTE: To compare results of failing tests, ",
            )
            lines.insert(
                12, "                             use idiff.py instead"
            )
            lines.insert(
                13,
                "  --data-files-used    Save a list of files used to a temporary file",
            )
            lines.insert(
                14, "  -m                   Create missing test results"
            )
            print("\n".join(lines))
    else:
        unittest.main()


def get_data_path(relative_path):
    """
    Return the absolute path to a data file when given the relative path
    as a string, or sequence of strings.

    """
    if not isinstance(relative_path, str):
        relative_path = os.path.join(*relative_path)
    test_data_dir = iris.config.TEST_DATA_DIR
    if test_data_dir is None:
        test_data_dir = ""
    data_path = os.path.join(test_data_dir, relative_path)

    if _EXPORT_DATAPATHS_FILE is not None:
        _EXPORT_DATAPATHS_FILE.write(data_path + "\n")

    if isinstance(data_path, str) and not os.path.exists(data_path):
        # if the file is gzipped, ungzip it and return the path of the ungzipped
        # file.
        gzipped_fname = data_path + ".gz"
        if os.path.exists(gzipped_fname):
            with gzip.open(gzipped_fname, "rb") as gz_fh:
                try:
                    with open(data_path, "wb") as fh:
                        fh.writelines(gz_fh)
                except IOError:
                    # Put ungzipped data file in a temporary path, since we
                    # can't write to the original path (maybe it is owned by
                    # the system.)
                    _, ext = os.path.splitext(data_path)
                    data_path = iris.util.create_temp_filename(suffix=ext)
                    with open(data_path, "wb") as fh:
                        fh.writelines(gz_fh)

    return data_path


class IrisTest_nometa(unittest.TestCase):
    """A subclass of unittest.TestCase which provides Iris specific testing functionality."""

    _assertion_counts = collections.defaultdict(int)

    @classmethod
    def setUpClass(cls):
        # Ensure that the CF profile if turned-off for testing.
        iris.site_configuration["cf_profile"] = None

    def _assert_str_same(
        self,
        reference_str,
        test_str,
        reference_filename,
        type_comparison_name="Strings",
    ):
        if reference_str != test_str:
            diff = "".join(
                difflib.unified_diff(
                    reference_str.splitlines(1),
                    test_str.splitlines(1),
                    "Reference",
                    "Test result",
                    "",
                    "",
                    0,
                )
            )
            self.fail(
                "%s do not match: %s\n%s"
                % (type_comparison_name, reference_filename, diff)
            )

    @staticmethod
    def get_result_path(relative_path):
        """
        Returns the absolute path to a result file when given the relative path
        as a string, or sequence of strings.

        """
        if not isinstance(relative_path, str):
            relative_path = os.path.join(*relative_path)
        return os.path.abspath(os.path.join(_RESULT_PATH, relative_path))

    def assertStringEqual(
        self, reference_str, test_str, type_comparison_name="strings"
    ):
        if reference_str != test_str:
            diff = "\n".join(
                difflib.unified_diff(
                    reference_str.splitlines(),
                    test_str.splitlines(),
                    "Reference",
                    "Test result",
                    "",
                    "",
                    0,
                )
            )
            self.fail(
                "{} do not match:\n{}".format(type_comparison_name, diff)
            )

    def result_path(self, basename=None, ext=""):
        """
        Return the full path to a test result, generated from the \
        calling file, class and, optionally, method.

        Optional kwargs :

            * basename    - File basename. If omitted, this is \
                            generated from the calling method.
            * ext         - Appended file extension.

        """
        if ext and not ext.startswith("."):
            ext = "." + ext

        # Generate the folder name from the calling file name.
        path = os.path.abspath(inspect.getfile(self.__class__))
        path = os.path.splitext(path)[0]
        sub_path = path.rsplit("iris", 1)[1].split("tests", 1)[1][1:]

        # Generate the file name from the calling function name?
        if basename is None:
            stack = inspect.stack()
            for frame in stack[1:]:
                if "test_" in frame[3]:
                    basename = frame[3].replace("test_", "")
                    break
        filename = basename + ext

        result = os.path.join(
            self.get_result_path(""),
            sub_path.replace("test_", ""),
            self.__class__.__name__.replace("Test_", ""),
            filename,
        )
        return result

    def assertCMLApproxData(self, cubes, reference_filename=None, **kwargs):
        # passes args and kwargs on to approx equal
        if isinstance(cubes, iris.cube.Cube):
            cubes = [cubes]
        if reference_filename is None:
            reference_filename = self.result_path(None, "cml")
            reference_filename = [self.get_result_path(reference_filename)]
        for i, cube in enumerate(cubes):
            fname = list(reference_filename)
            # don't want the ".cml" for the json stats file
            if fname[-1].endswith(".cml"):
                fname[-1] = fname[-1][:-4]
            fname[-1] += ".data.%d.json" % i
            self.assertDataAlmostEqual(cube.data, fname, **kwargs)
        self.assertCML(cubes, reference_filename, checksum=False)

    def assertCDL(self, netcdf_filename, reference_filename=None, flags="-h"):
        """
        Test that the CDL for the given netCDF file matches the contents
        of the reference file.

        If the environment variable IRIS_TEST_CREATE_MISSING is
        non-empty, the reference file is created if it doesn't exist.

        Args:

        * netcdf_filename:
            The path to the netCDF file.

        Kwargs:

        * reference_filename:
            The relative path (relative to the test results directory).
            If omitted, the result is generated from the calling
            method's name, class, and module using
            :meth:`iris.tests.IrisTest.result_path`.

        * flags:
            Command-line flags for `ncdump`, as either a whitespace
            separated string or an iterable. Defaults to '-h'.

        """
        if reference_filename is None:
            reference_path = self.result_path(None, "cdl")
        else:
            reference_path = self.get_result_path(reference_filename)

        # Convert the netCDF file to CDL file format.
        if flags is None:
            flags = []
        elif isinstance(flags, str):
            flags = flags.split()
        else:
            flags = list(map(str, flags))

        try:
            # Python3 only: use subprocess.run()
            args = ["ncdump"] + flags + [netcdf_filename]
            cdl = subprocess.check_output(args, stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError as exc:
            print(exc.output)
            raise

        # Ingest the CDL for comparison, excluding first line.
        lines = cdl.decode("ascii").splitlines()
        lines = lines[1:]

        # Ignore any lines of the general form "... :_NCProperties = ..."
        # (an extra global attribute, displayed by older versions of ncdump).
        re_ncprop = re.compile(r"^\s*:_NCProperties *=")
        lines = [line for line in lines if not re_ncprop.match(line)]

        # Sort the dimensions (except for the first, which can be unlimited).
        # This gives consistent CDL across different platforms.
        def sort_key(line):
            return ("UNLIMITED" not in line, line)

        dimension_lines = slice(
            lines.index("dimensions:") + 1, lines.index("variables:")
        )
        lines[dimension_lines] = sorted(lines[dimension_lines], key=sort_key)
        cdl = "\n".join(lines) + "\n"

        self._check_same(cdl, reference_path, type_comparison_name="CDL")

    def assertCML(self, cubes, reference_filename=None, checksum=True):
        """
        Test that the CML for the given cubes matches the contents of
        the reference file.

        If the environment variable IRIS_TEST_CREATE_MISSING is
        non-empty, the reference file is created if it doesn't exist.

        Args:

        * cubes:
            Either a Cube or a sequence of Cubes.

        Kwargs:

        * reference_filename:
            The relative path (relative to the test results directory).
            If omitted, the result is generated from the calling
            method's name, class, and module using
            :meth:`iris.tests.IrisTest.result_path`.

        * checksum:
            When True, causes the CML to include a checksum for each
            Cube's data. Defaults to True.

        """
        if isinstance(cubes, iris.cube.Cube):
            cubes = [cubes]
        if reference_filename is None:
            reference_filename = self.result_path(None, "cml")

        if isinstance(cubes, (list, tuple)):
            xml = iris.cube.CubeList(cubes).xml(
                checksum=checksum, order=False, byteorder=False
            )
        else:
            xml = cubes.xml(checksum=checksum, order=False, byteorder=False)
        reference_path = self.get_result_path(reference_filename)
        self._check_same(xml, reference_path)

    def assertTextFile(
        self, source_filename, reference_filename, desc="text file"
    ):
        """Check if two text files are the same, printing any diffs."""
        with open(source_filename) as source_file:
            source_text = source_file.readlines()
        with open(reference_filename) as reference_file:
            reference_text = reference_file.readlines()
        if reference_text != source_text:
            diff = "".join(
                difflib.unified_diff(
                    reference_text,
                    source_text,
                    "Reference",
                    "Test result",
                    "",
                    "",
                    0,
                )
            )
            self.fail(
                "%s does not match reference file: %s\n%s"
                % (desc, reference_filename, diff)
            )

    def assertDataAlmostEqual(self, data, reference_filename, **kwargs):
        reference_path = self.get_result_path(reference_filename)
        if self._check_reference_file(reference_path):
            kwargs.setdefault("err_msg", "Reference file %s" % reference_path)
            with open(reference_path, "r") as reference_file:
                stats = json.load(reference_file)
                self.assertEqual(stats.get("shape", []), list(data.shape))
                self.assertEqual(
                    stats.get("masked", False), ma.is_masked(data)
                )
                nstats = np.array(
                    (
                        stats.get("mean", 0.0),
                        stats.get("std", 0.0),
                        stats.get("max", 0.0),
                        stats.get("min", 0.0),
                    ),
                    dtype=np.float_,
                )
                if math.isnan(stats.get("mean", 0.0)):
                    self.assertTrue(math.isnan(data.mean()))
                else:
                    data_stats = np.array(
                        (data.mean(), data.std(), data.max(), data.min()),
                        dtype=np.float_,
                    )
                    self.assertArrayAllClose(nstats, data_stats, **kwargs)
        else:
            self._ensure_folder(reference_path)
            stats = collections.OrderedDict(
                [
                    ("std", np.float_(data.std())),
                    ("min", np.float_(data.min())),
                    ("max", np.float_(data.max())),
                    ("shape", data.shape),
                    ("masked", ma.is_masked(data)),
                    ("mean", np.float_(data.mean())),
                ]
            )
            with open(reference_path, "w") as reference_file:
                reference_file.write(json.dumps(stats))

    def assertFilesEqual(self, test_filename, reference_filename):
        reference_path = self.get_result_path(reference_filename)
        if self._check_reference_file(reference_path):
            fmt = "test file {!r} does not match reference {!r}."
            self.assertTrue(
                filecmp.cmp(test_filename, reference_path),
                fmt.format(test_filename, reference_path),
            )
        else:
            self._ensure_folder(reference_path)
            shutil.copy(test_filename, reference_path)

    def assertString(self, string, reference_filename=None):
        """
        Test that `string` matches the contents of the reference file.

        If the environment variable IRIS_TEST_CREATE_MISSING is
        non-empty, the reference file is created if it doesn't exist.

        Args:

        * string:
            The string to check.

        Kwargs:

        * reference_filename:
            The relative path (relative to the test results directory).
            If omitted, the result is generated from the calling
            method's name, class, and module using
            :meth:`iris.tests.IrisTest.result_path`.

        """
        if reference_filename is None:
            reference_path = self.result_path(None, "txt")
        else:
            reference_path = self.get_result_path(reference_filename)
        self._check_same(
            string, reference_path, type_comparison_name="Strings"
        )

    def assertRepr(self, obj, reference_filename):
        self.assertString(repr(obj), reference_filename)

    def _check_same(self, item, reference_path, type_comparison_name="CML"):
        if self._check_reference_file(reference_path):
            with open(reference_path, "rb") as reference_fh:
                reference = "".join(
                    part.decode("utf-8") for part in reference_fh.readlines()
                )
            self._assert_str_same(
                reference, item, reference_path, type_comparison_name
            )
        else:
            self._ensure_folder(reference_path)
            with open(reference_path, "wb") as reference_fh:
                reference_fh.writelines(part.encode("utf-8") for part in item)

    def assertXMLElement(self, obj, reference_filename):
        """
        Calls the xml_element method given obj and asserts the result is the same as the test file.

        """
        doc = xml.dom.minidom.Document()
        doc.appendChild(obj.xml_element(doc))
        # sort the attributes on xml elements before testing against known good state.
        # this is to be compatible with stored test output where xml attrs are stored in alphabetical order,
        # (which was default behaviour in python <3.8, but changed to insert order in >3.8)
        doc = iris.cube.Cube._sort_xml_attrs(doc)
        pretty_xml = doc.toprettyxml(indent="  ")
        reference_path = self.get_result_path(reference_filename)
        self._check_same(
            pretty_xml, reference_path, type_comparison_name="XML"
        )

    def assertArrayEqual(self, a, b, err_msg=""):
        np.testing.assert_array_equal(a, b, err_msg=err_msg)

    @contextlib.contextmanager
    def _recordWarningMatches(self, expected_regexp=""):
        # Record warnings raised matching a given expression.
        matches = []
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            yield matches
        messages = [str(warning.message) for warning in w]
        expr = re.compile(expected_regexp)
        matches.extend(message for message in messages if expr.search(message))

    @contextlib.contextmanager
    def assertWarnsRegexp(self, expected_regexp=""):
        # Check that a warning is raised matching a given expression.
        with self._recordWarningMatches(expected_regexp) as matches:
            yield

        msg = "Warning matching '{}' not raised."
        msg = msg.format(expected_regexp)
        self.assertTrue(matches, msg)

    @contextlib.contextmanager
    def assertNoWarningsRegexp(self, expected_regexp=""):
        # Check that no warning matching the given expression is raised.
        with self._recordWarningMatches(expected_regexp) as matches:
            yield

        msg = "Unexpected warning(s) raised, matching '{}' : {!r}."
        msg = msg.format(expected_regexp, matches)
        self.assertFalse(matches, msg)

    def _assertMaskedArray(self, assertion, a, b, strict, **kwargs):
        # Define helper function to extract unmasked values as a 1d
        # array.
        def unmasked_data_as_1d_array(array):
            array = ma.asarray(array)
            if array.ndim == 0:
                if array.mask:
                    data = np.array([])
                else:
                    data = np.array([array.data])
            else:
                data = array.data[~ma.getmaskarray(array)]
            return data

        # Compare masks. This will also check that the array shapes
        # match, which is not tested when comparing unmasked values if
        # strict is False.
        a_mask, b_mask = ma.getmaskarray(a), ma.getmaskarray(b)
        np.testing.assert_array_equal(a_mask, b_mask)

        if strict:
            assertion(a.data, b.data, **kwargs)
        else:
            assertion(
                unmasked_data_as_1d_array(a),
                unmasked_data_as_1d_array(b),
                **kwargs,
            )

    def assertMaskedArrayEqual(self, a, b, strict=False):
        """
        Check that masked arrays are equal. This requires the
        unmasked values and masks to be identical.

        Args:

        * a, b (array-like):
            Two arrays to compare.

        Kwargs:

        * strict (bool):
            If True, perform a complete mask and data array equality check.
            If False (default), the data array equality considers only unmasked
            elements.

        """
        self._assertMaskedArray(np.testing.assert_array_equal, a, b, strict)

    def assertArrayAlmostEqual(self, a, b, decimal=6):
        np.testing.assert_array_almost_equal(a, b, decimal=decimal)

    def assertMaskedArrayAlmostEqual(self, a, b, decimal=6, strict=False):
        """
        Check that masked arrays are almost equal. This requires the
        masks to be identical, and the unmasked values to be almost
        equal.

        Args:

        * a, b (array-like):
            Two arrays to compare.

        Kwargs:

        * strict (bool):
            If True, perform a complete mask and data array equality check.
            If False (default), the data array equality considers only unmasked
            elements.

        * decimal (int):
            Equality tolerance level for
            :meth:`numpy.testing.assert_array_almost_equal`, with the meaning
            'abs(desired-actual) < 0.5 * 10**(-decimal)'

        """
        self._assertMaskedArray(
            np.testing.assert_array_almost_equal, a, b, strict, decimal=decimal
        )

    def assertArrayAllClose(self, a, b, rtol=1.0e-7, atol=1.0e-8, **kwargs):
        """
        Check arrays are equal, within given relative + absolute tolerances.

        Args:

        * a, b (array-like):
            Two arrays to compare.

        Kwargs:

        * rtol, atol (float):
            Relative and absolute tolerances to apply.

        Any additional kwargs are passed to numpy.testing.assert_allclose.

        Performs pointwise toleranced comparison, and raises an assertion if
        the two are not equal 'near enough'.
        For full details see underlying routine numpy.allclose.

        """
        # Handle the 'err_msg' kwarg, which is the only API difference
        # between np.allclose and np.testing_assert_allclose.
        msg = kwargs.pop("err_msg", None)
        ok = np.allclose(a, b, rtol=rtol, atol=atol, **kwargs)
        if not ok:
            # Calculate errors above a pointwise tolerance : The method is
            # taken from "numpy.core.numeric.isclose".
            a, b = np.broadcast_arrays(a, b)
            errors = np.abs(a - b) - atol + rtol * np.abs(b)
            worst_inds = np.unravel_index(np.argmax(errors.flat), errors.shape)

            if msg is None:
                # Build a more useful message than np.testing.assert_allclose.
                msg = (
                    '\nARRAY CHECK FAILED "assertArrayAllClose" :'
                    "\n  with shapes={} {}, atol={}, rtol={}"
                    "\n  worst at element {} :  a={}  b={}"
                    "\n  absolute error ~{:.3g}, equivalent to rtol ~{:.3e}"
                )
                aval, bval = a[worst_inds], b[worst_inds]
                absdiff = np.abs(aval - bval)
                equiv_rtol = absdiff / bval
                msg = msg.format(
                    a.shape,
                    b.shape,
                    atol,
                    rtol,
                    worst_inds,
                    aval,
                    bval,
                    absdiff,
                    equiv_rtol,
                )

            raise AssertionError(msg)

    @contextlib.contextmanager
    def temp_filename(self, suffix=""):
        filename = iris.util.create_temp_filename(suffix)
        try:
            yield filename
        finally:
            os.remove(filename)

    def file_checksum(self, file_path):
        """
        Generate checksum from file.
        """
        with open(file_path, "rb") as in_file:
            return zlib.crc32(in_file.read())

    def _unique_id(self):
        """
        Returns the unique ID for the current assertion.

        The ID is composed of two parts: a unique ID for the current test
        (which is itself composed of the module, class, and test names), and
        a sequential counter (specific to the current test) that is incremented
        on each call.

        For example, calls from a "test_tx" routine followed by a "test_ty"
        routine might result in::
            test_plot.TestContourf.test_tx.0
            test_plot.TestContourf.test_tx.1
            test_plot.TestContourf.test_tx.2
            test_plot.TestContourf.test_ty.0

        """
        # Obtain a consistent ID for the current test.
        # NB. unittest.TestCase.id() returns different values depending on
        # whether the test has been run explicitly, or via test discovery.
        # For example:
        #   python tests/test_plot.py => '__main__.TestContourf.test_tx'
        #   ird -t => 'iris.tests.test_plot.TestContourf.test_tx'
        bits = self.id().split(".")
        if bits[0] == "__main__":
            floc = sys.modules["__main__"].__file__
            path, file_name = os.path.split(os.path.abspath(floc))
            bits[0] = os.path.splitext(file_name)[0]
            folder, location = os.path.split(path)
            bits = [location] + bits
            while location not in ["iris", "gallery_tests"]:
                folder, location = os.path.split(folder)
                bits = [location] + bits
        test_id = ".".join(bits)

        # Derive the sequential assertion ID within the test
        assertion_id = self._assertion_counts[test_id]
        self._assertion_counts[test_id] += 1

        return test_id + "." + str(assertion_id)

    def _check_reference_file(self, reference_path):
        reference_exists = os.path.isfile(reference_path)
        if not (
            reference_exists or os.environ.get("IRIS_TEST_CREATE_MISSING")
        ):
            msg = "Missing test result: {}".format(reference_path)
            raise AssertionError(msg)
        return reference_exists

    def _ensure_folder(self, path):
        dir_path = os.path.dirname(path)
        if not os.path.exists(dir_path):
            os.makedirs(dir_path)

    def check_graphic(self):
        """
        Check the hash of the current matplotlib figure matches the expected
        image hash for the current graphic test.

        To create missing image test results, set the IRIS_TEST_CREATE_MISSING
        environment variable before running the tests. This will result in new
        and appropriately "<hash>.png" image files being generated in the image
        output directory, and the imagerepo.json file being updated.

        """
        import imagehash
        from PIL import Image

        dev_mode = os.environ.get("IRIS_TEST_CREATE_MISSING")
        unique_id = self._unique_id()
        repo_fname = os.path.join(_RESULT_PATH, "imagerepo.json")
        with open(repo_fname, "rb") as fi:
            repo = json.load(codecs.getreader("utf-8")(fi))

        try:
            #: The path where the images generated by the tests should go.
            image_output_directory = os.path.join(
                os.path.dirname(__file__), "result_image_comparison"
            )
            if not os.access(image_output_directory, os.W_OK):
                if not os.access(os.getcwd(), os.W_OK):
                    raise IOError(
                        "Write access to a local disk is required "
                        "to run image tests.  Run the tests from a "
                        "current working directory you have write "
                        "access to to avoid this issue."
                    )
                else:
                    image_output_directory = os.path.join(
                        os.getcwd(), "iris_image_test_output"
                    )
            result_fname = os.path.join(
                image_output_directory, "result-" + unique_id + ".png"
            )

            if not os.path.isdir(image_output_directory):
                # Handle race-condition where the directories are
                # created sometime between the check above and the
                # creation attempt below.
                try:
                    os.makedirs(image_output_directory)
                except OSError as err:
                    # Don't care about "File exists"
                    if err.errno != 17:
                        raise

            def _create_missing():
                fname = "{}.png".format(phash)
                base_uri = (
                    "https://scitools.github.io/test-iris-imagehash/"
                    "images/v4/{}"
                )
                uri = base_uri.format(fname)
                hash_fname = os.path.join(image_output_directory, fname)
                uris = repo.setdefault(unique_id, [])
                uris.append(uri)
                print("Creating image file: {}".format(hash_fname))
                figure.savefig(hash_fname)
                msg = "Creating imagerepo entry: {} -> {}"
                print(msg.format(unique_id, uri))
                lock = filelock.FileLock(
                    os.path.join(_RESULT_PATH, "imagerepo.lock")
                )
                # The imagerepo.json file is a critical resource, so ensure
                # thread safe read/write behaviour via platform independent
                # file locking.
                with lock.acquire(timeout=600):
                    with open(repo_fname, "wb") as fo:
                        json.dump(
                            repo,
                            codecs.getwriter("utf-8")(fo),
                            indent=4,
                            sort_keys=True,
                        )

            # Calculate the test result perceptual image hash.
            buffer = io.BytesIO()
            figure = plt.gcf()
            figure.savefig(buffer, format="png")
            buffer.seek(0)
            phash = imagehash.phash(Image.open(buffer), hash_size=_HASH_SIZE)

            if unique_id not in repo:
                if dev_mode:
                    _create_missing()
                else:
                    figure.savefig(result_fname)
                    emsg = "Missing image test result: {}."
                    raise AssertionError(emsg.format(unique_id))
            else:
                uris = repo[unique_id]
                # Extract the hex basename strings from the uris.
                hexes = [
                    os.path.splitext(os.path.basename(uri))[0] for uri in uris
                ]
                # Create the expected perceptual image hashes from the uris.
                to_hash = imagehash.hex_to_hash
                expected = [to_hash(uri_hex) for uri_hex in hexes]

                # Calculate hamming distance vector for the result hash.
                distances = [e - phash for e in expected]

                if np.all([hd > _HAMMING_DISTANCE for hd in distances]):
                    if dev_mode:
                        _create_missing()
                    else:
                        figure.savefig(result_fname)
                        msg = (
                            "Bad phash {} with hamming distance {} "
                            "for test {}."
                        )
                        msg = msg.format(phash, distances, unique_id)
                        if _DISPLAY_FIGURES:
                            emsg = "Image comparison would have failed: {}"
                            print(emsg.format(msg))
                        else:
                            emsg = "Image comparison failed: {}"
                            raise AssertionError(emsg.format(msg))

            if _DISPLAY_FIGURES:
                plt.show()

        finally:
            plt.close()

    def _remove_testcase_patches(self):
        """Helper to remove per-testcase patches installed by :meth:`patch`."""
        # Remove all patches made, ignoring errors.
        for p in self.testcase_patches:
            p.stop()
        # Reset per-test patch control variable.
        self.testcase_patches.clear()

    def patch(self, *args, **kwargs):
        """
        Install a mock.patch, to be removed after the current test.

        The patch is created with mock.patch(*args, **kwargs).

        Returns:
            The substitute object returned by patch.start().

        For example::

            mock_call = self.patch('module.Class.call', return_value=1)
            module_Class_instance.call(3, 4)
            self.assertEqual(mock_call.call_args_list, [mock.call(3, 4)])

        """
        # Make the new patch and start it.
        patch = mock.patch(*args, **kwargs)
        start_result = patch.start()

        # Create the per-testcases control variable if it does not exist.
        # NOTE: this mimics a setUp method, but continues to work when a
        # subclass defines its own setUp.
        if not hasattr(self, "testcase_patches"):
            self.testcase_patches = {}

        # When installing the first patch, schedule remove-all at cleanup.
        if not self.testcase_patches:
            self.addCleanup(self._remove_testcase_patches)

        # Record the new patch and start object for reference.
        self.testcase_patches[patch] = start_result

        # Return patch replacement object.
        return start_result

    def assertArrayShapeStats(self, result, shape, mean, std_dev, rtol=1e-6):
        """
        Assert that the result, a cube, has the provided shape and that the
        mean and standard deviation of the data array are also as provided.
        Thus build confidence that a cube processing operation, such as a
        cube.regrid, has maintained its behaviour.

        """
        self.assertEqual(result.shape, shape)
        self.assertArrayAllClose(result.data.mean(), mean, rtol=rtol)
        self.assertArrayAllClose(result.data.std(), std_dev, rtol=rtol)

    def assertDictEqual(self, lhs, rhs, msg=None):
        """
        This method overrides unittest.TestCase.assertDictEqual (new in Python3.1)
        in order to cope with dictionary comparison where the value of a key may
        be a numpy array.

        """
        if not isinstance(lhs, Mapping):
            emsg = (
                f"Provided LHS argument is not a 'Mapping', got {type(lhs)}."
            )
            self.fail(emsg)

        if not isinstance(rhs, Mapping):
            emsg = (
                f"Provided RHS argument is not a 'Mapping', got {type(rhs)}."
            )
            self.fail(emsg)

        if set(lhs.keys()) != set(rhs.keys()):
            emsg = f"{lhs!r} != {rhs!r}."
            self.fail(emsg)

        for key in lhs.keys():
            lvalue, rvalue = lhs[key], rhs[key]

            if ma.isMaskedArray(lvalue) or ma.isMaskedArray(rvalue):
                if not ma.isMaskedArray(lvalue):
                    emsg = (
                        f"Dictionary key {key!r} values are not equal, "
                        f"the LHS value has type {type(lvalue)} and "
                        f"the RHS value has type {ma.core.MaskedArray}."
                    )
                    raise AssertionError(emsg)

                if not ma.isMaskedArray(rvalue):
                    emsg = (
                        f"Dictionary key {key!r} values are not equal, "
                        f"the LHS value has type {ma.core.MaskedArray} and "
                        f"the RHS value has type {type(lvalue)}."
                    )
                    raise AssertionError(emsg)

                self.assertMaskedArrayEqual(lvalue, rvalue)
            elif isinstance(lvalue, np.ndarray) or isinstance(
                rvalue, np.ndarray
            ):
                if not isinstance(lvalue, np.ndarray):
                    emsg = (
                        f"Dictionary key {key!r} values are not equal, "
                        f"the LHS value has type {type(lvalue)} and "
                        f"the RHS value has type {np.ndarray}."
                    )
                    raise AssertionError(emsg)

                if not isinstance(rvalue, np.ndarray):
                    emsg = (
                        f"Dictionary key {key!r} values are not equal, "
                        f"the LHS value has type {np.ndarray} and "
                        f"the RHS value has type {type(rvalue)}."
                    )
                    raise AssertionError(emsg)

                self.assertArrayEqual(lvalue, rvalue)
            else:
                if lvalue != rvalue:
                    emsg = (
                        f"Dictionary key {key!r} values are not equal, "
                        f"{lvalue!r} != {rvalue!r}."
                    )
                    raise AssertionError(emsg)

    def assertEqualAndKind(self, value, expected):
        # Check a value, and also its type 'kind' = float/integer/string.
        self.assertEqual(value, expected)
        self.assertEqual(
            np.array(value).dtype.kind, np.array(expected).dtype.kind
        )


# An environment variable controls whether test timings are output.
#
# NOTE: to run tests with timing output, nosetests cannot be used.
# At present, that includes not using "python setup.py test"
# The typically best way is like this :
#    $ export IRIS_TEST_TIMINGS=1
#    $ python -m unittest discover -s iris.tests
# and commonly adding ...
#    | grep "TIMING TEST" >iris_test_output.txt
#
_PRINT_TEST_TIMINGS = bool(int(os.environ.get("IRIS_TEST_TIMINGS", 0)))


def _method_path(meth, cls):
    return ".".join([cls.__module__, cls.__name__, meth.__name__])


def _testfunction_timing_decorator(fn, cls):
    # Function decorator for making a testcase print its execution time.
    @functools.wraps(fn)
    def inner(*args, **kwargs):
        start_time = datetime.datetime.now()
        try:
            result = fn(*args, **kwargs)
        finally:
            end_time = datetime.datetime.now()
            elapsed_time = (end_time - start_time).total_seconds()
            msg = '\n  TEST TIMING -- "{}" took : {:12.6f} sec.'
            name = _method_path(fn, cls)
            print(msg.format(name, elapsed_time))
        return result

    return inner


def iristest_timing_decorator(cls):
    # Class decorator to make all "test_.." functions print execution timings.
    if _PRINT_TEST_TIMINGS:
        # NOTE: 'dir' scans *all* class properties, including inherited ones.
        attr_names = dir(cls)
        for attr_name in attr_names:
            attr = getattr(cls, attr_name)
            if callable(attr) and attr_name.startswith("test"):
                attr = _testfunction_timing_decorator(attr, cls)
                setattr(cls, attr_name, attr)
    return cls


class _TestTimingsMetaclass(type):
    # An alternative metaclass for IrisTest subclasses, which makes
    # them print execution timings for all the testcases.
    # This is equivalent to applying the @iristest_timing_decorator to
    # every test class that inherits from IrisTest.
    # NOTE: however, it means you *cannot* specify a different metaclass for
    # your test class inheriting from IrisTest.
    # See below for how to solve that where needed.
    def __new__(cls, clsname, base_classes, attrs):
        result = type.__new__(cls, clsname, base_classes, attrs)
        if _PRINT_TEST_TIMINGS:
            result = iristest_timing_decorator(result)
        return result


class IrisTest(IrisTest_nometa, metaclass=_TestTimingsMetaclass):
    # Derive the 'ordinary' IrisTest from IrisTest_nometa, but add the
    # metaclass that enables test timings output.
    # This means that all subclasses also get the timing behaviour.
    # However, if a different metaclass is *wanted* for an IrisTest subclass,
    # this would cause a metaclass conflict.
    # Instead, you can inherit from IrisTest_nometa and apply the
    # @iristest_timing_decorator explicitly to your new testclass.
    pass


get_result_path = IrisTest.get_result_path


class GraphicsTestMixin:

    # nose directive: dispatch tests concurrently.
    _multiprocess_can_split_ = True

    def setUp(self):
        # Acquire threading non re-entrant blocking lock to ensure
        # thread-safe plotting.
        _lock.acquire()
        # Make sure we have no unclosed plots from previous tests before
        # generating this one.
        if MPL_AVAILABLE:
            plt.close("all")

    def tearDown(self):
        # If a plotting test bombs out it can leave the current figure
        # in an odd state, so we make sure it's been disposed of.
        if MPL_AVAILABLE:
            plt.close("all")
        # Release the non re-entrant blocking lock.
        _lock.release()


class GraphicsTest(GraphicsTestMixin, IrisTest):
    pass


class GraphicsTest_nometa(GraphicsTestMixin, IrisTest_nometa):
    # Graphicstest without the metaclass providing test timings.
    pass


def skip_data(fn):
    """
    Decorator to choose whether to run tests, based on the availability of
    external data.

    Example usage:
        @skip_data
        class MyDataTests(tests.IrisTest):
            ...

    """
    no_data = (
        not iris.config.TEST_DATA_DIR
        or not os.path.isdir(iris.config.TEST_DATA_DIR)
        or os.environ.get("IRIS_TEST_NO_DATA")
    )

    skip = unittest.skipIf(
        condition=no_data, reason="Test(s) require external data."
    )

    return skip(fn)


def skip_gdal(fn):
    """
    Decorator to choose whether to run tests, based on the availability of the
    GDAL library.

    Example usage:
        @skip_gdal
        class MyGeoTiffTests(test.IrisTest):
            ...

    """
    skip = unittest.skipIf(
        condition=not GDAL_AVAILABLE, reason="Test requires 'gdal'."
    )
    return skip(fn)


def skip_plot(fn):
    """
    Decorator to choose whether to run tests, based on the availability of the
    matplotlib library.

    Example usage:
        @skip_plot
        class MyPlotTests(test.GraphicsTest):
            ...

    """
    skip = unittest.skipIf(
        condition=not MPL_AVAILABLE,
        reason="Graphics tests require the matplotlib library.",
    )

    return skip(fn)


skip_sample_data = unittest.skipIf(
    not SAMPLE_DATA_AVAILABLE,
    ('Test(s) require "iris-sample-data", ' "which is not available."),
)


skip_nc_time_axis = unittest.skipIf(
    not NC_TIME_AXIS_AVAILABLE,
    'Test(s) require "nc_time_axis", which is not available.',
)


skip_inet = unittest.skipIf(
    not INET_AVAILABLE,
    ('Test(s) require an "internet connection", ' "which is not available."),
)


skip_stratify = unittest.skipIf(
    not STRATIFY_AVAILABLE,
    'Test(s) require "python-stratify", which is not available.',
)


def no_warnings(func):
    """
    Provides a decorator to ensure that there are no warnings raised
    within the test, otherwise the test will fail.

    """

    @functools.wraps(func)
    def wrapped(self, *args, **kwargs):
        with mock.patch("warnings.warn") as warn:
            result = func(self, *args, **kwargs)
        self.assertEqual(
            0,
            warn.call_count,
            ("Got unexpected warnings." " \n{}".format(warn.call_args_list)),
        )
        return result

    return wrapped



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.

"""Unit tests for the `iris.analysis.cartography.area_weights` function"""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests
import iris.tests.stock as stock
import iris.analysis.cartography


class TestInvalidUnits(tests.IrisTest):
    def test_latitude_no_units(self):
        cube = stock.lat_lon_cube()
        cube.coord("longitude").guess_bounds()
        cube.coord("latitude").guess_bounds()
        cube.coord("latitude").units = None
        with self.assertRaisesRegex(
            ValueError, "Units of degrees or " "radians required"
        ):
            iris.analysis.cartography.area_weights(cube)

    def test_longitude_no_units(self):
        cube = stock.lat_lon_cube()
        cube.coord("latitude").guess_bounds()
        cube.coord("longitude").guess_bounds()
        cube.coord("longitude").units = None
        with self.assertRaisesRegex(
            ValueError, "Units of degrees or " "radians required"
        ):
            iris.analysis.cartography.area_weights(cube)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Test function :func:`iris.util.find_discontiguities"""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

import numpy as np

from iris.tests.stock import simple_3d
from iris.tests.stock import sample_2d_latlons
from iris.tests.stock import make_bounds_discontiguous_at_point

from iris.util import find_discontiguities


def full2d_global():
    return sample_2d_latlons(transformed=True)


@tests.skip_data
class Test(tests.IrisTest):
    def setUp(self):
        # Set up a 2d lat-lon cube with 2d coordinates that have been
        # transformed so they are not in a regular lat-lon grid.
        # Then generate a discontiguity at a single lat-lon point.
        self.testcube_discontig = full2d_global()
        make_bounds_discontiguous_at_point(self.testcube_discontig, 3, 3)
        # Repeat that for a discontiguity in the grid 'Y' direction.
        self.testcube_discontig_along_y = full2d_global()
        make_bounds_discontiguous_at_point(
            self.testcube_discontig_along_y, 2, 4, in_y=True
        )

    def test_find_discontiguities(self):
        # Check that the mask we generate when making the discontiguity
        # matches that generated by find_discontiguities
        cube = self.testcube_discontig
        expected = cube.data.mask
        returned = find_discontiguities(cube)
        self.assertTrue(np.all(expected == returned))

    def test_find_discontiguities_in_y(self):
        # Check that the mask we generate when making the discontiguity
        # matches that generated by find_discontiguities
        cube = self.testcube_discontig_along_y
        expected = cube.data.mask
        returned = find_discontiguities(cube)
        self.assertTrue(np.all(expected == returned))

    def test_find_discontiguities_1d_coord(self):
        # Check that an error is raised when we try and use
        # find_discontiguities on 1D coordinates:
        cube = simple_3d()
        with self.assertRaises(NotImplementedError):
            find_discontiguities(cube)

    def test_find_discontiguities_with_atol(self):
        cube = self.testcube_discontig
        # Choose a very large absolute tolerance which will result in fine
        # discontiguities being disregarded
        atol = 100
        # Construct an array the size of the points array filled with 'False'
        # to represent a mask showing no discontiguities
        expected = np.zeros(cube.shape, dtype=bool)
        returned = find_discontiguities(cube, abs_tol=atol)
        self.assertTrue(np.all(expected == returned))

    def test_find_discontiguities_with_rtol(self):
        cube = self.testcube_discontig
        # Choose a very large relative tolerance which will result in fine
        # discontiguities being disregarded
        rtol = 1000
        # Construct an array the size of the points array filled with 'False'
        # to represent a mask showing no discontiguities
        expected = np.zeros(cube.shape, dtype=bool)
        returned = find_discontiguities(cube, rel_tol=rtol)
        self.assertTrue(np.all(expected == returned))


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Test function :func:`iris.util._slice_data_with_keys`.

Note: much of the functionality really belongs to the other routines,
:func:`iris.util._build_full_slice_given_keys`, and
:func:`column_slices_generator`.
However, it is relatively simple to test multiple aspects of all three here
in combination.

"""

# import iris tests first so that some things can be initialised before
# importing anything else
import iris.tests as tests

import numpy as np

from iris.util import _slice_data_with_keys
from iris._lazy_data import as_lazy_data, as_concrete_data


class DummyArray:
    # A dummy array-like that records the keys of indexing calls.
    def __init__(self, shape, _indexing_record_list=None):
        self.shape = shape
        self.ndim = len(shape)
        if _indexing_record_list is None:
            _indexing_record_list = []
        self._getitem_call_keys = _indexing_record_list

    def __getitem__(self, keys):
        # Add the indexing keys to the call list.
        self._getitem_call_keys.append(keys)
        # Return a new object with the correct derived shape, and record its
        # indexing operations in the same key list as this.
        shape_array = np.zeros(self.shape)
        shape_array = shape_array.__getitem__(keys)
        new_shape = shape_array.shape
        return DummyArray(
            new_shape, _indexing_record_list=self._getitem_call_keys
        )


class Indexer:
    # An object to make __getitem__ arglists from indexing operations.
    def __getitem__(self, keys):
        return keys


# An Indexer object for generating indexing keys in a nice visible way.
Index = Indexer()


class MixinIndexingTest:
    def check(self, shape, keys, expect_call_keys=None, expect_map=None):
        data = DummyArray(shape)
        dim_map, _ = _slice_data_with_keys(data, keys)
        if expect_call_keys is not None:
            calls_got = data._getitem_call_keys
            # Check that the indexing keys applied were the expected ones.
            equal = len(calls_got) == len(expect_call_keys)
            for act_call, expect_call in zip(calls_got, expect_call_keys):
                equal &= len(act_call) == len(expect_call)
                # A problem here is that in each call, some keys may be
                # *arrays*, and arrays can't be compared in the "normal"
                # way.  So we must use np.all for comparison :-(
                for act_key, expect_key in zip(act_call, expect_call):
                    equal &= np.asanyarray(act_key).dtype == np.asanyarray(
                        expect_key
                    ).dtype and np.all(act_key == expect_key)
            errmsg = "Different key lists:\n{!s}\n!=\n{!s}\n"

            def showkeys(keys_list):
                msg = "[\n  "
                msg += "\n  ".join(str(x) for x in keys_list)
                msg += "\n]"
                return msg

            self.assertTrue(
                equal,
                errmsg.format(showkeys(calls_got), showkeys(expect_call_keys)),
            )
        if expect_map is not None:
            self.assertEqual(dim_map, expect_map)


class Test_indexing(MixinIndexingTest, tests.IrisTest):
    # Check the indexing operations performed for various requested keys.

    def test_0d_nokeys(self):
        # Performs *no* underlying indexing operation.
        self.check((), Index[()], [])

    def test_1d_int(self):
        self.check((4,), Index[2], [(2,)])

    def test_1d_all(self):
        self.check((3,), Index[:], [(slice(None),)])

    def test_1d_tuple(self):
        # The call makes tuples into 1-D arrays, and a trailing Ellipsis is
        # added (for the 1-D case only).
        self.check(
            (3,), Index[((2, 0, 1),)], [(np.array([2, 0, 1]), Ellipsis)]
        )

    def test_fail_1d_2keys(self):
        msg = "More slices .* than dimensions"
        with self.assertRaisesRegex(IndexError, msg):
            self.check((3,), Index[1, 2])

    def test_fail_empty_slice(self):
        msg = "Cannot index with zero length slice"
        with self.assertRaisesRegex(IndexError, msg):
            self.check((3,), Index[1:1])

    def test_2d_tuple(self):
        # Like the above, but there is an extra no-op at the start and no
        # trailing Ellipsis is generated.
        self.check(
            (3, 2),
            Index[((2, 0, 1),)],
            [(slice(None), slice(None)), (np.array([2, 0, 1]), slice(None))],
        )

    def test_2d_two_tuples(self):
        # Could be treated as fancy indexing, but must not be !
        # Two separate 2-D indexing operations.
        self.check(
            (3, 2),
            Index[(2, 0, 1, 1), (0, 1, 0, 1)],
            [
                (np.array([2, 0, 1, 1]), slice(None)),
                (slice(None), np.array([0, 1, 0, 1])),
            ],
        )

    def test_2d_tuple_and_value(self):
        # The two keys are applied in separate operations, and in the reverse
        # order (?) :  The second op is then slicing a 1-D array, not 2-D.
        self.check(
            (3, 5),
            Index[(2, 0, 1), 3],
            [(slice(None), 3), (np.array([2, 0, 1]), Ellipsis)],
        )

    def test_2d_single_int(self):
        self.check((3, 4), Index[2], [(2, slice(None))])

    def test_2d_multiple_int(self):
        self.check((3, 4), Index[2, 1:3], [(2, slice(1, 3))])

    def test_3d_1int(self):
        self.check((3, 4, 5), Index[2], [(2, slice(None), slice(None))])

    def test_3d_2int(self):
        self.check((3, 4, 5), Index[2, 3], [(2, 3, slice(None))])

    def test_3d_tuple_and_value(self):
        # The two keys are applied in separate operations, and in the reverse
        # order (?) : The second op is slicing a 2-D array, not 3-D.
        self.check(
            (3, 5, 7),
            Index[(2, 0, 1), 4],
            [
                (slice(None), 4, slice(None)),
                (np.array([2, 0, 1]), slice(None)),
            ],
        )

    def test_3d_ellipsis_last(self):
        self.check((3, 4, 5), Index[2, ...], [(2, slice(None), slice(None))])

    def test_3d_ellipsis_first_1int(self):
        self.check((3, 4, 5), Index[..., 2], [(slice(None), slice(None), 2)])

    def test_3d_ellipsis_first_2int(self):
        self.check((3, 4, 5), Index[..., 2, 3], [(slice(None), 2, 3)])

    def test_3d_multiple_tuples(self):
        # Where there are TWO or more tuple keys, this could be misinterpreted
        # as 'fancy' indexing :  It should resolve into multiple calls.
        self.check(
            (3, 4, 5),
            Index[(1, 2, 1), :, (2, 2, 3)],
            [
                (slice(None), slice(None), slice(None)),
                (np.array([1, 2, 1]), slice(None), slice(None)),
                (slice(None), slice(None), np.array([2, 2, 3])),
            ],
        )
        # NOTE: there seem to be an extra initial [:, :, :].
        # That's just what it does at present.


class Test_dimensions_mapping(MixinIndexingTest, tests.IrisTest):
    # Check the dimensions map returned for various requested keys.

    def test_1d_nochange(self):
        self.check((3,), Index[1:2], expect_map={None: None, 0: 0})

    def test_1d_1int_losedim0(self):
        self.check((3,), Index[1], expect_map={None: None, 0: None})

    def test_1d_tuple_nochange(self):
        # A selection index leaves the dimension intact.
        self.check((3,), Index[((1, 0, 1, 2),)], expect_map={None: None, 0: 0})

    def test_1d_1tuple_nochange(self):
        # A selection index with only one value in it *still* leaves the
        # dimension intact.
        self.check((3,), Index[((2,),)], expect_map={None: None, 0: 0})

    def test_1d_slice_nochange(self):
        # A slice leaves the dimension intact.
        self.check((3,), Index[1:7], expect_map={None: None, 0: 0})

    def test_2d_nochange(self):
        self.check((3, 4), Index[:, :], expect_map={None: None, 0: 0, 1: 1})

    def test_2d_losedim0(self):
        self.check((3, 4), Index[1, :], expect_map={None: None, 0: None, 1: 0})

    def test_2d_losedim1(self):
        self.check(
            (3, 4), Index[1:4, 2], expect_map={None: None, 0: 0, 1: None}
        )

    def test_2d_loseboth(self):
        # Two indices give scalar result.
        self.check(
            (3, 4), Index[1, 2], expect_map={None: None, 0: None, 1: None}
        )

    def test_3d_losedim1(self):
        # Cutting out the middle dim.
        self.check(
            (3, 4, 2),
            Index[:, 2],
            expect_map={None: None, 0: 0, 1: None, 2: 1},
        )


class TestResults(tests.IrisTest):
    # Integration-style test, exercising (mostly) the same cases as above,
    # but checking actual results, for both real and lazy array inputs.

    def check(self, real_data, keys, expect_result, expect_map):
        real_data = np.array(real_data)
        lazy_data = as_lazy_data(real_data, real_data.shape)
        real_dim_map, real_result = _slice_data_with_keys(real_data, keys)
        lazy_dim_map, lazy_result = _slice_data_with_keys(lazy_data, keys)
        lazy_result = as_concrete_data(lazy_result)
        self.assertArrayEqual(real_result, expect_result)
        self.assertArrayEqual(lazy_result, expect_result)
        self.assertEqual(real_dim_map, expect_map)
        self.assertEqual(lazy_dim_map, expect_map)

    def test_1d_int(self):
        self.check([1, 2, 3, 4], Index[2], [3], {None: None, 0: None})

    def test_1d_all(self):
        self.check([1, 2, 3], Index[:], [1, 2, 3], {None: None, 0: 0})

    def test_1d_tuple(self):
        self.check(
            [1, 2, 3], Index[((2, 0, 1, 0),)], [3, 1, 2, 1], {None: None, 0: 0}
        )

    def test_fail_1d_2keys(self):
        msg = "More slices .* than dimensions"
        with self.assertRaisesRegex(IndexError, msg):
            self.check([1, 2, 3], Index[1, 2], None, None)

    def test_fail_empty_slice(self):
        msg = "Cannot index with zero length slice"
        with self.assertRaisesRegex(IndexError, msg):
            self.check([1, 2, 3], Index[1:1], None, None)

    def test_2d_tuple(self):
        self.check(
            [[11, 12], [21, 22], [31, 32]],
            Index[((2, 0, 1),)],
            [[31, 32], [11, 12], [21, 22]],
            {None: None, 0: 0, 1: 1},
        )

    def test_2d_two_tuples(self):
        # Could be treated as fancy indexing, but must not be !
        # Two separate 2-D indexing operations.
        self.check(
            [[11, 12, 13], [21, 22, 23], [31, 32, 33]],
            Index[(2, 0), (0, 1, 0, 1)],
            [[31, 32, 31, 32], [11, 12, 11, 12]],
            {None: None, 0: 0, 1: 1},
        )

    def test_2d_tuple_and_value(self):
        # The two keys are applied in separate operations, and in the reverse
        # order (?) :  The second op is then slicing a 1-D array, not 2-D.
        self.check(
            [[11, 12, 13, 14], [21, 22, 23, 24], [31, 32, 33, 34]],
            Index[(2, 0, 1), 3],
            [34, 14, 24],
            {None: None, 0: 0, 1: None},
        )

    def test_2d_single_int(self):
        self.check(
            [[11, 12, 13], [21, 22, 23], [31, 32, 33]],
            Index[1],
            [21, 22, 23],
            {None: None, 0: None, 1: 0},
        )

    def test_2d_int_slice(self):
        self.check(
            [[11, 12, 13], [21, 22, 23], [31, 32, 33]],
            Index[2, 1:3],
            [32, 33],
            {None: None, 0: None, 1: 0},
        )

    def test_3d_1int(self):
        self.check(
            [
                [[111, 112, 113], [121, 122, 123]],
                [[211, 212, 213], [221, 222, 223]],
                [[311, 312, 313], [321, 322, 323]],
            ],
            Index[1],
            [[211, 212, 213], [221, 222, 223]],
            {None: None, 0: None, 1: 0, 2: 1},
        )

    def test_3d_2int(self):
        self.check(
            [
                [[111, 112, 113], [121, 122, 123], [131, 132, 133]],
                [[211, 212, 213], [221, 222, 223], [231, 232, 233]],
            ],
            Index[1, 2],
            [231, 232, 233],
            {None: None, 0: None, 1: None, 2: 0},
        )

    def test_3d_tuple_and_value(self):
        # The two keys are applied in separate operations, and in the reverse
        # order (?) : The second op is slicing a 2-D array, not 3-D.
        self.check(
            [
                [[111, 112, 113, 114], [121, 122, 123, 124]],
                [[211, 212, 213, 214], [221, 222, 223, 224]],
                [[311, 312, 313, 314], [321, 322, 323, 324]],
            ],
            Index[(2, 0, 1), 1],
            [[321, 322, 323, 324], [121, 122, 123, 124], [221, 222, 223, 224]],
            {None: None, 0: 0, 1: None, 2: 1},
        )

    def test_3d_ellipsis_last(self):
        self.check(
            [
                [[111, 112, 113], [121, 122, 123]],
                [[211, 212, 213], [221, 222, 223]],
                [[311, 312, 313], [321, 322, 323]],
            ],
            Index[2, ...],
            [[311, 312, 313], [321, 322, 323]],
            {None: None, 0: None, 1: 0, 2: 1},
        )

    def test_3d_ellipsis_first_1int(self):
        self.check(
            [
                [[111, 112, 113, 114], [121, 122, 123, 124]],
                [[211, 212, 213, 214], [221, 222, 223, 224]],
                [[311, 312, 313, 314], [321, 322, 323, 324]],
            ],
            Index[..., 2],
            [[113, 123], [213, 223], [313, 323]],
            {None: None, 0: 0, 1: 1, 2: None},
        )

    def test_3d_ellipsis_mid_1int(self):
        self.check(
            [
                [[111, 112, 113], [121, 122, 123]],
                [[211, 212, 213], [221, 222, 223]],
                [[311, 312, 313], [321, 322, 323]],
            ],
            Index[..., 1, ...],
            [[121, 122, 123], [221, 222, 223], [321, 322, 323]],
            {None: None, 0: 0, 1: None, 2: 1},
        )

    def test_3d_ellipsis_first_2int(self):
        self.check(
            [
                [[111, 112, 113], [121, 122, 123]],
                [[211, 212, 213], [221, 222, 223]],
                [[311, 312, 313], [321, 322, 323]],
            ],
            Index[..., 1, 2],
            [123, 223, 323],
            {None: None, 0: 0, 1: None, 2: None},
        )

    def test_3d_multiple_tuples(self):
        # Where there are TWO or more tuple keys, this could be misinterpreted
        # as 'fancy' indexing :  It should resolve into multiple calls.
        self.check(
            [
                [[111, 112, 113, 114], [121, 122, 123, 124]],
                [[211, 212, 213, 214], [221, 222, 223, 224]],
                [[311, 312, 313, 314], [321, 322, 323, 324]],
            ],
            Index[(1, 2, 1), :, (2, 2, 3)],
            [
                [[213, 213, 214], [223, 223, 224]],
                [[313, 313, 314], [323, 323, 324]],
                [[213, 213, 214], [223, 223, 224]],
            ],
            {None: None, 0: 0, 1: 1, 2: 2},
        )
        # NOTE: there seem to be an extra initial [:, :, :].
        # That's just what it does at present.


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Unit tests for :func:`iris.fileformats.pp_load_rules._dim_or_aux`."""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

from iris.coords import DimCoord, AuxCoord
from iris.fileformats.pp_load_rules import _dim_or_aux


class Test(tests.IrisTest):
    def setUp(self):
        self.mono = list(range(5))
        self.non_mono = [0, 1, 3, 2, 4]
        self.std_name = "depth"
        self.units = "m"
        self.attr = {"positive": "up", "wibble": "wobble"}

    def test_dim_monotonic(self):
        result = _dim_or_aux(
            self.mono,
            standard_name=self.std_name,
            units=self.units,
            attributes=self.attr.copy(),
        )
        expected = DimCoord(
            self.mono,
            standard_name=self.std_name,
            units=self.units,
            attributes=self.attr,
        )
        self.assertEqual(result, expected)

    def test_dim_non_monotonic(self):
        result = _dim_or_aux(
            self.non_mono,
            standard_name=self.std_name,
            units=self.units,
            attributes=self.attr.copy(),
        )
        attr = self.attr.copy()
        del attr["positive"]
        expected = AuxCoord(
            self.non_mono,
            standard_name=self.std_name,
            units=self.units,
            attributes=attr,
        )
        self.assertEqual(result, expected)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Test plot of time coord with non-gregorian calendar.

"""

# import iris tests first so that some things can be initialised before
# importing anything else
import iris.tests as tests

import cftime
import numpy as np

from iris.coords import AuxCoord

from cf_units import Unit

if tests.NC_TIME_AXIS_AVAILABLE:
    from nc_time_axis import CalendarDateTime


# Run tests in no graphics mode if matplotlib is not available.
if tests.MPL_AVAILABLE:
    import iris.plot as iplt


@tests.skip_nc_time_axis
@tests.skip_plot
class Test(tests.GraphicsTest):
    def test_360_day_calendar(self):
        n = 360
        calendar = "360_day"
        time_unit = Unit("days since 1970-01-01 00:00", calendar=calendar)
        time_coord = AuxCoord(np.arange(n), "time", units=time_unit)
        times = [time_unit.num2date(point) for point in time_coord.points]
        times = [
            cftime.datetime(
                atime.year,
                atime.month,
                atime.day,
                atime.hour,
                atime.minute,
                atime.second,
            )
            for atime in times
        ]
        expected_ydata = np.array(
            [CalendarDateTime(time, calendar) for time in times]
        )
        (line1,) = iplt.plot(time_coord)
        result_ydata = line1.get_ydata()
        self.assertArrayEqual(expected_ydata, result_ydata)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.


"""
This system test module is useful to identify if some of the key components required for Iris are available.

The system tests can be run with ``python setup.py test --system-tests``.

"""

# import iris tests first so that some things can be initialised before importing anything else

import cf_units
import numpy as np

import iris
import iris.tests as tests


class SystemInitialTest(tests.IrisTest):
    def system_test_supported_filetypes(self):
        nx, ny = 60, 60
        data = np.arange(nx * ny, dtype=">f4").reshape(nx, ny)

        laty = np.linspace(0, 59, ny).astype("f8")
        lonx = np.linspace(30, 89, nx).astype("f8")

        def horiz_cs():
            return iris.coord_systems.GeogCS(6371229)

        cm = iris.cube.Cube(data, "wind_speed", units="m s-1")
        cm.add_dim_coord(
            iris.coords.DimCoord(
                laty, "latitude", units="degrees", coord_system=horiz_cs()
            ),
            0,
        )
        cm.add_dim_coord(
            iris.coords.DimCoord(
                lonx, "longitude", units="degrees", coord_system=horiz_cs()
            ),
            1,
        )
        cm.add_aux_coord(
            iris.coords.AuxCoord(
                np.array([9], "i8"), "forecast_period", units="hours"
            )
        )
        hours_since_epoch = cf_units.Unit(
            "hours since epoch", cf_units.CALENDAR_GREGORIAN
        )
        cm.add_aux_coord(
            iris.coords.AuxCoord(
                np.array([3], "i8"), "time", units=hours_since_epoch
            )
        )
        cm.add_aux_coord(
            iris.coords.AuxCoord(
                np.array([99], "i8"), long_name="pressure", units="Pa"
            )
        )

        filetypes = (".nc", ".pp")
        for filetype in filetypes:
            saved_tmpfile = iris.util.create_temp_filename(suffix=filetype)
            iris.save(cm, saved_tmpfile)

            new_cube = iris.load_cube(saved_tmpfile)
            self.assertCML(
                new_cube, ("system", "supported_filetype_%s.cml" % filetype)
            )

    def system_test_imports_general(self):
        if tests.MPL_AVAILABLE:
            import matplotlib  # noqa
        import netCDF4  # noqa


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Unit tests for the :class:`iris.common.mixin.LimitedAttributeDict`.

"""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

from unittest import mock
import numpy as np

from iris.common.mixin import LimitedAttributeDict


class Test(tests.IrisTest):
    def setUp(self):
        self.forbidden_keys = LimitedAttributeDict._forbidden_keys
        self.emsg = "{!r} is not a permitted attribute"

    def test__invalid_keys(self):
        for key in self.forbidden_keys:
            with self.assertRaisesRegex(ValueError, self.emsg.format(key)):
                _ = LimitedAttributeDict(**{key: None})

    def test___eq__(self):
        values = dict(
            one=mock.sentinel.one,
            two=mock.sentinel.two,
            three=mock.sentinel.three,
        )
        left = LimitedAttributeDict(**values)
        right = LimitedAttributeDict(**values)
        self.assertEqual(left, right)
        self.assertEqual(left, values)

    def test___eq___numpy(self):
        values = dict(one=np.arange(1), two=np.arange(2), three=np.arange(3))
        left = LimitedAttributeDict(**values)
        right = LimitedAttributeDict(**values)
        self.assertEqual(left, right)
        self.assertEqual(left, values)
        values = dict(one=np.arange(1), two=np.arange(1), three=np.arange(1))
        left = LimitedAttributeDict(dict(one=0, two=0, three=0))
        right = LimitedAttributeDict(**values)
        self.assertEqual(left, right)
        self.assertEqual(left, values)

    def test___setitem__(self):
        for key in self.forbidden_keys:
            item = LimitedAttributeDict()
            with self.assertRaisesRegex(ValueError, self.emsg.format(key)):
                item[key] = None

    def test_update(self):
        for key in self.forbidden_keys:
            item = LimitedAttributeDict()
            with self.assertRaisesRegex(ValueError, self.emsg.format(key)):
                other = {key: None}
                item.update(other)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Test the io/__init__.py module.

"""

# import iris tests first so that some things can be initialised before importing anything else
import iris.tests as tests

from io import BytesIO

import iris.fileformats as iff
import iris.io


class TestDecodeUri(tests.IrisTest):
    def test_decode_uri(self):
        tests = {
            "/data/local/someDir/PP/COLPEX/COLPEX_16a_pj001.pp": (
                "file",
                "/data/local/someDir/PP/COLPEX/COLPEX_16a_pj001.pp",
            ),
            r"C:\data\local\someDir\PP\COLPEX\COLPEX_16a_pj001.pp": (
                "file",
                r"C:\data\local\someDir\PP\COLPEX\COLPEX_16a_pj001.pp",
            ),
            "file:///data/local/someDir/PP/COLPEX/COLPEX_16a_pj001.pp": (
                "file",
                "///data/local/someDir/PP/COLPEX/COLPEX_16a_pj001.pp",
            ),
            "http://www.somehost.com:8080/resource/thing.grib": (
                "http",
                "//www.somehost.com:8080/resource/thing.grib",
            ),
            "/data/local/someDir/2013-11-25T13:49:17.632797": (
                "file",
                "/data/local/someDir/2013-11-25T13:49:17.632797",
            ),
        }
        for uri, pair in tests.items():
            self.assertEqual(pair, iris.io.decode_uri(uri))


class TestFileFormatPicker(tests.IrisTest):
    def test_known_formats(self):
        self.assertString(
            str(iff.FORMAT_AGENT),
            tests.get_result_path(("file_load", "known_loaders.txt")),
        )

    @tests.skip_data
    def test_format_picker(self):
        # ways to test the format picker = list of (format-name, file-spec)
        test_specs = [
            (
                "NetCDF",
                ["NetCDF", "global", "xyt", "SMALL_total_column_co2.nc"],
            ),
            (
                "NetCDF 64 bit offset format",
                ["NetCDF", "global", "xyt", "SMALL_total_column_co2.nc.k2"],
            ),
            (
                "NetCDF_v4",
                ["NetCDF", "global", "xyt", "SMALL_total_column_co2.nc4.k3"],
            ),
            (
                "NetCDF_v4",
                ["NetCDF", "global", "xyt", "SMALL_total_column_co2.nc4.k4"],
            ),
            ("UM Fieldsfile (FF) post v5.2", ["FF", "n48_multi_field"]),
            (
                "GRIB",
                ["GRIB", "grib1_second_order_packing", "GRIB_00008_FRANX01"],
            ),
            ("GRIB", ["GRIB", "jpeg2000", "file.grib2"]),
            ("UM Post Processing file (PP)", ["PP", "simple_pp", "global.pp"]),
            (
                "UM Post Processing file (PP) little-endian",
                ["PP", "little_endian", "qrparm.orog.pp"],
            ),
            (
                "UM Fieldsfile (FF) ancillary",
                ["FF", "ancillary_fixed_length_header"],
            ),
            #            ('BUFR',
            #                ['BUFR', 'mss', 'BUFR_Samples',
            #                 'JUPV78_EGRR_121200_00002501']),
            (
                "NIMROD",
                [
                    "NIMROD",
                    "uk2km",
                    "WO0000000003452",
                    "201007020900_u1096_ng_ey00_visibility0180_screen_2km",
                ],
            ),
            #            ('NAME',
            #                ['NAME', '20100509_18Z_variablesource_12Z_VAAC',
            #                 'Fields_grid1_201005110000.txt']),
        ]

        # test that each filespec is identified as the expected format
        for (expected_format_name, file_spec) in test_specs:
            test_path = tests.get_data_path(file_spec)
            with open(test_path, "rb") as test_file:
                a = iff.FORMAT_AGENT.get_spec(test_path, test_file)
                self.assertEqual(a.name, expected_format_name)

    def test_format_picker_nodata(self):
        # The following is to replace the above at some point as no real files
        # are required.
        # (Used binascii.unhexlify() to convert from hex to binary)

        # Packaged grib, magic number offset by set length, this length is
        # specific to WMO bulletin headers
        header_lengths = [21, 80, 41, 42]
        for header_length in header_lengths:
            binary_string = header_length * b"\x00" + b"GRIB" + b"\x00" * 100
            with BytesIO(b"rw") as bh:
                bh.write(binary_string)
                bh.name = "fake_file_handle"
                a = iff.FORMAT_AGENT.get_spec(bh.name, bh)
            self.assertEqual(a.name, "GRIB")

    def test_open_dap(self):
        # tests that *ANY* http or https URL is seen as an OPeNDAP service.
        # This may need to change in the future if other protocols are
        # supported.
        DAP_URI = "http://geoport.whoi.edu/thredds/dodsC/bathy/gom15"
        a = iff.FORMAT_AGENT.get_spec(DAP_URI, None)
        self.assertEqual(a.name, "NetCDF OPeNDAP")


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Test function :func:`iris.fileformats._pyke_rules.compiled_krb.\
fc_rules_cf_fc.build_dimension_coordinate`.

"""

# import iris tests first so that some things can be initialised before
# importing anything else
import iris.tests as tests

from unittest import mock
import warnings

import numpy as np

from iris.coords import AuxCoord, DimCoord
from iris.fileformats._pyke_rules.compiled_krb.fc_rules_cf_fc import \
    build_dimension_coordinate


class RulesTestMixin:
    def setUp(self):
        # Create dummy pyke engine.
        self.engine = mock.Mock(
            cube=mock.Mock(),
            cf_var=mock.Mock(dimensions=('foo', 'bar')),
            filename='DUMMY',
            cube_parts=dict(coordinates=[]))

        # Create patch for deferred loading that prevents attempted
        # file access. This assumes that self.cf_coord_var and
        # self.cf_bounds_var are defined in the test case.
        def patched__getitem__(proxy_self, keys):
            for var in (self.cf_coord_var, self.cf_bounds_var):
                if proxy_self.variable_name == var.cf_name:
                    return var[keys]
            raise RuntimeError()

        self.deferred_load_patch = mock.patch(
            'iris.fileformats.netcdf.NetCDFDataProxy.__getitem__',
            new=patched__getitem__)

        # Patch the helper function that retrieves the bounds cf variable.
        # This avoids the need for setting up further mocking of cf objects.
        self.use_climatology_bounds = False  # Set this when you need to.

        def get_cf_bounds_var(coord_var):
            return self.cf_bounds_var, self.use_climatology_bounds

        self.get_cf_bounds_var_patch = mock.patch(
            'iris.fileformats._pyke_rules.compiled_krb.'
            'fc_rules_cf_fc.get_cf_bounds_var',
            new=get_cf_bounds_var)


class TestCoordConstruction(tests.IrisTest, RulesTestMixin):
    def setUp(self):
        # Call parent setUp explicitly, because of how unittests work.
        RulesTestMixin.setUp(self)

        bounds = np.arange(12).reshape(6, 2)
        self.cf_bounds_var = mock.Mock(
            dimensions=('x', 'nv'),
            cf_name='wibble_bnds',
            shape=bounds.shape,
            __getitem__=lambda self, key: bounds[key])
        self.bounds = bounds

    def _set_cf_coord_var(self, points):
        self.cf_coord_var = mock.Mock(
            dimensions=('foo',),
            cf_name='wibble',
            cf_data=mock.Mock(spec=[]),
            standard_name=None,
            long_name='wibble',
            units='days since 1970-01-01',
            calendar=None,
            shape=points.shape,
            dtype=points.dtype,
            __getitem__=lambda self, key: points[key])

    def check_case_dim_coord_construction(self, climatology=False):
        # Test a generic dimension coordinate, with or without
        # a climatological coord.
        self.use_climatology_bounds = climatology
        self._set_cf_coord_var(np.arange(6))

        expected_coord = DimCoord(
            self.cf_coord_var[:],
            long_name=self.cf_coord_var.long_name,
            var_name=self.cf_coord_var.cf_name,
            units=self.cf_coord_var.units,
            bounds=self.bounds,
            climatological=climatology)

        # Asserts must lie within context manager because of deferred loading.
        with self.deferred_load_patch, self.get_cf_bounds_var_patch:
            build_dimension_coordinate(self.engine, self.cf_coord_var)

            # Test that expected coord is built and added to cube.
            self.engine.cube.add_dim_coord.assert_called_with(
                expected_coord, [0])

    def test_dim_coord_construction(self):
        self.check_case_dim_coord_construction(climatology=False)

    def test_dim_coord_construction__climatology(self):
        self.check_case_dim_coord_construction(climatology=True)

    def test_dim_coord_construction_masked_array(self):
        self._set_cf_coord_var(np.ma.array(
            np.arange(6),
            mask=[True, False, False, False, False, False],
            fill_value=-999,
        ))

        expected_coord = DimCoord(
            np.array([-999, 1, 2, 3, 4, 5]),
            long_name=self.cf_coord_var.long_name,
            var_name=self.cf_coord_var.cf_name,
            units=self.cf_coord_var.units,
            bounds=self.bounds)

        with warnings.catch_warnings(record=True) as w:
            # Asserts must lie within context manager because of deferred
            # loading.
            with self.deferred_load_patch, self.get_cf_bounds_var_patch:
                build_dimension_coordinate(self.engine, self.cf_coord_var)

                # Test that expected coord is built and added to cube.
                self.engine.cube.add_dim_coord.assert_called_with(
                    expected_coord, [0])

            # Assert warning is raised
            assert len(w) == 1
            assert 'Gracefully filling' in w[0].message.args[0]

    def test_dim_coord_construction_masked_array_mask_does_nothing(self):
        self._set_cf_coord_var(np.ma.array(
            np.arange(6),
            mask=False,
        ))

        expected_coord = DimCoord(
            self.cf_coord_var[:],
            long_name=self.cf_coord_var.long_name,
            var_name=self.cf_coord_var.cf_name,
            units=self.cf_coord_var.units,
            bounds=self.bounds)

        with warnings.catch_warnings(record=True) as w:
            # Asserts must lie within context manager because of deferred
            # loading.
            with self.deferred_load_patch, self.get_cf_bounds_var_patch:
                build_dimension_coordinate(self.engine, self.cf_coord_var)

                # Test that expected coord is built and added to cube.
                self.engine.cube.add_dim_coord.assert_called_with(
                    expected_coord, [0])

            # Assert no warning is raised
            assert len(w) == 0

    def test_dim_coord_construction_masked_bounds_mask_does_nothing(self):
        self.bounds = np.ma.array(np.arange(12).reshape(6, 2), mask=False)
        self._set_cf_coord_var(np.arange(6))

        expected_coord = DimCoord(
            self.cf_coord_var[:],
            long_name=self.cf_coord_var.long_name,
            var_name=self.cf_coord_var.cf_name,
            units=self.cf_coord_var.units,
            bounds=self.bounds)

        with warnings.catch_warnings(record=True) as w:
            # Asserts must lie within context manager because of deferred
            # loading.
            with self.deferred_load_patch, self.get_cf_bounds_var_patch:
                build_dimension_coordinate(self.engine, self.cf_coord_var)

                # Test that expected coord is built and added to cube.
                self.engine.cube.add_dim_coord.assert_called_with(
                    expected_coord, [0])

            # Assert no warning is raised
            assert len(w) == 0

    def test_aux_coord_construction(self):
        # Use non monotonically increasing coordinates to force aux coord
        # construction.
        self._set_cf_coord_var(np.array([1, 3, 2, 4, 6, 5]))

        expected_coord = AuxCoord(
            self.cf_coord_var[:],
            long_name=self.cf_coord_var.long_name,
            var_name=self.cf_coord_var.cf_name,
            units=self.cf_coord_var.units,
            bounds=self.bounds)

        warning_patch = mock.patch('warnings.warn')

        # Asserts must lie within context manager because of deferred loading.
        with warning_patch, self.deferred_load_patch, \
                self.get_cf_bounds_var_patch:
            build_dimension_coordinate(self.engine, self.cf_coord_var)

            # Test that expected coord is built and added to cube.
            self.engine.cube.add_aux_coord.assert_called_with(
                expected_coord, [0])
            self.assertIn("creating 'wibble' auxiliary coordinate instead",
                          warnings.warn.call_args[0][0])


class TestBoundsVertexDim(tests.IrisTest, RulesTestMixin):
    def setUp(self):
        # Call parent setUp explicitly, because of how unittests work.
        RulesTestMixin.setUp(self)
        # Create test coordinate cf variable.
        points = np.arange(6)
        self.cf_coord_var = mock.Mock(
            dimensions=('foo',),
            cf_name='wibble',
            standard_name=None,
            long_name='wibble',
            cf_data=mock.Mock(spec=[]),
            units='m',
            shape=points.shape,
            dtype=points.dtype,
            __getitem__=lambda self, key: points[key])

    def test_slowest_varying_vertex_dim(self):
        # Create the bounds cf variable.
        bounds = np.arange(12).reshape(2, 6)
        self.cf_bounds_var = mock.Mock(
            dimensions=('nv', 'foo'),
            cf_name='wibble_bnds',
            shape=bounds.shape,
            __getitem__=lambda self, key: bounds[key])

        # Expected bounds on the resulting coordinate should be rolled so that
        # the vertex dimension is at the end.
        expected_bounds = bounds.transpose()
        expected_coord = DimCoord(
            self.cf_coord_var[:],
            long_name=self.cf_coord_var.long_name,
            var_name=self.cf_coord_var.cf_name,
            units=self.cf_coord_var.units,
            bounds=expected_bounds)

        # Asserts must lie within context manager because of deferred loading.
        with self.deferred_load_patch, self.get_cf_bounds_var_patch:
            build_dimension_coordinate(self.engine, self.cf_coord_var)

            # Test that expected coord is built and added to cube.
            self.engine.cube.add_dim_coord.assert_called_with(
                expected_coord, [0])

            # Test that engine.cube_parts container is correctly populated.
            expected_list = [(expected_coord, self.cf_coord_var.cf_name)]
            self.assertEqual(self.engine.cube_parts['coordinates'],
                             expected_list)

    def test_fastest_varying_vertex_dim(self):
        bounds = np.arange(12).reshape(6, 2)
        self.cf_bounds_var = mock.Mock(
            dimensions=('foo', 'nv'),
            cf_name='wibble_bnds',
            shape=bounds.shape,
            __getitem__=lambda self, key: bounds[key])

        expected_coord = DimCoord(
            self.cf_coord_var[:],
            long_name=self.cf_coord_var.long_name,
            var_name=self.cf_coord_var.cf_name,
            units=self.cf_coord_var.units,
            bounds=bounds)

        # Asserts must lie within context manager because of deferred loading.
        with self.deferred_load_patch, self.get_cf_bounds_var_patch:
            build_dimension_coordinate(self.engine, self.cf_coord_var)

            # Test that expected coord is built and added to cube.
            self.engine.cube.add_dim_coord.assert_called_with(
                expected_coord, [0])

            # Test that engine.cube_parts container is correctly populated.
            expected_list = [(expected_coord, self.cf_coord_var.cf_name)]
            self.assertEqual(self.engine.cube_parts['coordinates'],
                             expected_list)

    def test_fastest_with_different_dim_names(self):
        # Despite the dimension names 'x' differing from the coord's
        # which is 'foo' (as permitted by the cf spec),
        # this should still work because the vertex dim is the fastest varying.
        bounds = np.arange(12).reshape(6, 2)
        self.cf_bounds_var = mock.Mock(
            dimensions=('x', 'nv'),
            cf_name='wibble_bnds',
            shape=bounds.shape,
            __getitem__=lambda self, key: bounds[key])

        expected_coord = DimCoord(
            self.cf_coord_var[:],
            long_name=self.cf_coord_var.long_name,
            var_name=self.cf_coord_var.cf_name,
            units=self.cf_coord_var.units,
            bounds=bounds)

        # Asserts must lie within context manager because of deferred loading.
        with self.deferred_load_patch, self.get_cf_bounds_var_patch:
            build_dimension_coordinate(self.engine, self.cf_coord_var)

            # Test that expected coord is built and added to cube.
            self.engine.cube.add_dim_coord.assert_called_with(
                expected_coord, [0])

            # Test that engine.cube_parts container is correctly populated.
            expected_list = [(expected_coord, self.cf_coord_var.cf_name)]
            self.assertEqual(self.engine.cube_parts['coordinates'],
                             expected_list)


class TestCircular(tests.IrisTest, RulesTestMixin):
    # Test the rules logic for marking a coordinate "circular".
    def setUp(self):
        # Call parent setUp explicitly, because of how unittests work.
        RulesTestMixin.setUp(self)
        self.cf_bounds_var = None

    def _make_vars(self, points, bounds=None, units='degrees'):
        points = np.array(points)
        self.cf_coord_var = mock.MagicMock(
            dimensions=('foo',),
            cf_name='wibble',
            standard_name=None,
            long_name='wibble',
            cf_data=mock.Mock(spec=[]),
            units=units,
            shape=points.shape,
            dtype=points.dtype,
            __getitem__=lambda self, key: points[key])
        if bounds:
            bounds = np.array(bounds).reshape(
                self.cf_coord_var.shape + (2,))
            self.cf_bounds_var = mock.Mock(
                dimensions=('x', 'nv'),
                cf_name='wibble_bnds',
                shape=bounds.shape,
                __getitem__=lambda self, key: bounds[key])

    def _check_circular(self, circular, *args, **kwargs):
        if 'coord_name' in kwargs:
            coord_name = kwargs.pop('coord_name')
        else:
            coord_name = 'longitude'
        self._make_vars(*args, **kwargs)
        with self.deferred_load_patch, self.get_cf_bounds_var_patch:
            build_dimension_coordinate(self.engine, self.cf_coord_var,
                                       coord_name=coord_name)
            self.assertEqual(self.engine.cube.add_dim_coord.call_count, 1)
            coord, dims = self.engine.cube.add_dim_coord.call_args[0]
        self.assertEqual(coord.circular, circular)

    def check_circular(self, *args, **kwargs):
        self._check_circular(True, *args, **kwargs)

    def check_noncircular(self, *args, **kwargs):
        self._check_circular(False, *args, **kwargs)

    def test_single_zero_noncircular(self):
        self.check_noncircular([0.0])

    def test_single_lt_modulus_noncircular(self):
        self.check_noncircular([-1.0])

    def test_single_eq_modulus_circular(self):
        self.check_circular([360.0])

    def test_single_gt_modulus_circular(self):
        self.check_circular([361.0])

    def test_single_bounded_noncircular(self):
        self.check_noncircular([180.0], bounds=[90.0, 240.0])

    def test_single_bounded_circular(self):
        self.check_circular([180.0], bounds=[90.0, 450.0])

    def test_multiple_unbounded_circular(self):
        self.check_circular([0.0, 90.0, 180.0, 270.0])

    def test_non_angle_noncircular(self):
        points = [0.0, 90.0, 180.0, 270.0]
        self.check_noncircular(points, units='m')

    def test_non_longitude_noncircular(self):
        points = [0.0, 90.0, 180.0, 270.0]
        self.check_noncircular(points, coord_name='depth')

    def test_multiple_unbounded_irregular_noncircular(self):
        self.check_noncircular([0.0, 90.0, 189.999, 270.0])

    def test_multiple_unbounded_offset_circular(self):
        self.check_circular([45.0, 135.0, 225.0, 315.0])

    def test_multiple_unbounded_shortrange_circular(self):
        self.check_circular([0.0, 90.0, 180.0, 269.9999])

    def test_multiple_bounded_circular(self):
        self.check_circular([0.0, 120.3, 240.0],
                            bounds=[[-45.0, 50.0],
                                    [100.0, 175.0],
                                    [200.0, 315.0]])

    def test_multiple_bounded_noncircular(self):
        self.check_noncircular([0.0, 120.3, 240.0],
                               bounds=[[-45.0, 50.0],
                                       [100.0, 175.0],
                                       [200.0, 355.0]])


class TestCircularScalar(tests.IrisTest, RulesTestMixin):
    def setUp(self):
        RulesTestMixin.setUp(self)

    def _make_vars(self, bounds):
        # Create cf vars for the coordinate and its bounds.
        # Note that for a scalar the shape of the array from
        # the cf var is (), rather than (1,).
        points = np.array([0.])
        self.cf_coord_var = mock.Mock(
            dimensions=(),
            cf_name='wibble',
            standard_name=None,
            long_name='wibble',
            units='degrees',
            cf_data=mock.Mock(spec=[]),
            shape=(),
            dtype=points.dtype,
            __getitem__=lambda self, key: points[key])

        bounds = np.array(bounds)
        self.cf_bounds_var = mock.Mock(
            dimensions=(u'bnds'),
            cf_name='wibble_bnds',
            shape=bounds.shape,
            __getitem__=lambda self, key: bounds[key])

    def _assert_circular(self, value):
        with self.deferred_load_patch, self.get_cf_bounds_var_patch:
            build_dimension_coordinate(self.engine, self.cf_coord_var,
                                       coord_name='longitude')
            self.assertEqual(self.engine.cube.add_aux_coord.call_count, 1)
            coord, dims = self.engine.cube.add_aux_coord.call_args[0]
        self.assertEqual(coord.circular, value)

    def test_two_bounds_noncircular(self):
        self._make_vars([0., 180.])
        self._assert_circular(False)

    def test_two_bounds_circular(self):
        self._make_vars([0., 360.])
        self._assert_circular(True)

    def test_two_bounds_circular_decreasing(self):
        self._make_vars([360., 0.])
        self._assert_circular(True)

    def test_two_bounds_circular_alt(self):
        self._make_vars([-180., 180.])
        self._assert_circular(True)

    def test_two_bounds_circular_alt_decreasing(self):
        self._make_vars([180., -180.])
        self._assert_circular(True)

    def test_four_bounds(self):
        self._make_vars([0., 10., 20., 30.])
        self._assert_circular(False)


if __name__ == '__main__':
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Unit tests for the :class:`iris.coords.AuxCoord` class.

Note: a lot of these methods are actually defined by the :class:`Coord` class,
but can only be tested on concrete instances (DimCoord or AuxCoord).

"""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

import numpy as np
import numpy.ma as ma

from iris.tests.unit.coords import (
    CoordTestMixin,
    lazyness_string,
    coords_all_dtypes_and_lazynesses,
)

from cf_units import Unit
from iris.coords import AuxCoord
from iris._lazy_data import as_lazy_data


class AuxCoordTestMixin(CoordTestMixin):
    # Define a 2-D default array shape.
    def setupTestArrays(self, shape=(2, 3), masked=False):
        super().setupTestArrays(shape, masked=masked)


class Test__init__(tests.IrisTest, AuxCoordTestMixin):
    # Test for AuxCoord creation, with various combinations of points and
    # bounds = real / lazy / None.
    def setUp(self):
        self.setupTestArrays(masked=True)

    def test_lazyness_and_dtype_combinations(self):
        for (
            coord,
            points_type_name,
            bounds_type_name,
        ) in coords_all_dtypes_and_lazynesses(self, AuxCoord):
            pts = coord.core_points()
            bds = coord.core_bounds()
            # Check properties of points.
            if points_type_name == "real":
                # Real points.
                if coord.dtype == self.pts_real.dtype:
                    self.assertArraysShareData(
                        pts,
                        self.pts_real,
                        "Points are not the same data as the provided array.",
                    )
                    self.assertIsNot(
                        pts,
                        self.pts_real,
                        "Points array is the same instance as the provided "
                        "array.",
                    )
                else:
                    # the original points were cast to a test dtype.
                    check_pts = self.pts_real.astype(coord.dtype)
                    self.assertEqualRealArraysAndDtypes(pts, check_pts)
            else:
                # Lazy points : the core data may be promoted to float.
                check_pts = self.pts_lazy.astype(pts.dtype)
                self.assertEqualLazyArraysAndDtypes(pts, check_pts)
                # The realisation type should be correct, though.
                target_dtype = coord.dtype
                self.assertEqual(coord.points.dtype, target_dtype)

            # Check properties of bounds.
            if bounds_type_name == "real":
                # Real bounds.
                if coord.bounds_dtype == self.bds_real.dtype:
                    self.assertArraysShareData(
                        bds,
                        self.bds_real,
                        "Bounds are not the same data as the provided array.",
                    )
                    self.assertIsNot(
                        pts,
                        self.pts_real,
                        "Bounds array is the same instance as the provided "
                        "array.",
                    )
                else:
                    # the original bounds were cast to a test dtype.
                    check_bds = self.bds_real.astype(coord.bounds_dtype)
                    self.assertEqualRealArraysAndDtypes(bds, check_bds)
            elif bounds_type_name == "lazy":
                # Lazy points : the core data may be promoted to float.
                check_bds = self.bds_lazy.astype(bds.dtype)
                self.assertEqualLazyArraysAndDtypes(bds, check_bds)
                # The realisation type should be correct, though.
                target_dtype = coord.bounds_dtype
                self.assertEqual(coord.bounds.dtype, target_dtype)

    def test_fail_bounds_shape_mismatch(self):
        bds_shape = list(self.bds_real.shape)
        bds_shape[0] += 1
        bds_wrong = np.zeros(bds_shape)
        msg = "Bounds shape must be compatible with points shape"
        with self.assertRaisesRegex(ValueError, msg):
            AuxCoord(self.pts_real, bounds=bds_wrong)

    def test_no_masked_pts_real(self):
        data = self.no_masked_pts_real
        self.assertTrue(ma.isMaskedArray(data))
        self.assertEqual(ma.count_masked(data), 0)
        coord = AuxCoord(data)
        self.assertFalse(coord.has_lazy_points())
        self.assertTrue(ma.isMaskedArray(coord.points))
        self.assertEqual(ma.count_masked(coord.points), 0)

    def test_no_masked_pts_lazy(self):
        data = self.no_masked_pts_lazy
        computed = data.compute()
        self.assertTrue(ma.isMaskedArray(computed))
        self.assertEqual(ma.count_masked(computed), 0)
        coord = AuxCoord(data)
        self.assertTrue(coord.has_lazy_points())
        self.assertTrue(ma.isMaskedArray(coord.points))
        self.assertEqual(ma.count_masked(coord.points), 0)

    def test_masked_pts_real(self):
        data = self.masked_pts_real
        self.assertTrue(ma.isMaskedArray(data))
        self.assertTrue(ma.count_masked(data))
        coord = AuxCoord(data)
        self.assertFalse(coord.has_lazy_points())
        self.assertTrue(ma.isMaskedArray(coord.points))
        self.assertTrue(ma.count_masked(coord.points))

    def test_masked_pts_lazy(self):
        data = self.masked_pts_lazy
        computed = data.compute()
        self.assertTrue(ma.isMaskedArray(computed))
        self.assertTrue(ma.count_masked(computed))
        coord = AuxCoord(data)
        self.assertTrue(coord.has_lazy_points())
        self.assertTrue(ma.isMaskedArray(coord.points))
        self.assertTrue(ma.count_masked(coord.points))

    def test_no_masked_bds_real(self):
        data = self.no_masked_bds_real
        self.assertTrue(ma.isMaskedArray(data))
        self.assertEqual(ma.count_masked(data), 0)
        coord = AuxCoord(self.pts_real, bounds=data)
        self.assertFalse(coord.has_lazy_bounds())
        self.assertTrue(ma.isMaskedArray(coord.bounds))
        self.assertEqual(ma.count_masked(coord.bounds), 0)

    def test_no_masked_bds_lazy(self):
        data = self.no_masked_bds_lazy
        computed = data.compute()
        self.assertTrue(ma.isMaskedArray(computed))
        self.assertEqual(ma.count_masked(computed), 0)
        coord = AuxCoord(self.pts_real, bounds=data)
        self.assertTrue(coord.has_lazy_bounds())
        self.assertTrue(ma.isMaskedArray(coord.bounds))
        self.assertEqual(ma.count_masked(coord.bounds), 0)

    def test_masked_bds_real(self):
        data = self.masked_bds_real
        self.assertTrue(ma.isMaskedArray(data))
        self.assertTrue(ma.count_masked(data))
        coord = AuxCoord(self.pts_real, bounds=data)
        self.assertFalse(coord.has_lazy_bounds())
        self.assertTrue(ma.isMaskedArray(coord.bounds))
        self.assertTrue(ma.count_masked(coord.bounds))

    def test_masked_bds_lazy(self):
        data = self.masked_bds_lazy
        computed = data.compute()
        self.assertTrue(ma.isMaskedArray(computed))
        self.assertTrue(ma.count_masked(computed))
        coord = AuxCoord(self.pts_real, bounds=data)
        self.assertTrue(coord.has_lazy_bounds())
        self.assertTrue(ma.isMaskedArray(coord.bounds))
        self.assertTrue(ma.count_masked(coord.bounds))


class Test_core_points(tests.IrisTest, AuxCoordTestMixin):
    # Test for AuxCoord.core_points() with various types of points and bounds.
    def setUp(self):
        self.setupTestArrays()

    def test_real_points(self):
        coord = AuxCoord(self.pts_real)
        result = coord.core_points()
        self.assertArraysShareData(
            result,
            self.pts_real,
            "core_points() do not share data with the internal array.",
        )

    def test_lazy_points(self):
        coord = AuxCoord(self.pts_lazy)
        result = coord.core_points()
        self.assertEqualLazyArraysAndDtypes(result, self.pts_lazy)

    def test_lazy_points_realise(self):
        coord = AuxCoord(self.pts_lazy)
        real_points = coord.points
        result = coord.core_points()
        self.assertEqualRealArraysAndDtypes(result, real_points)


class Test_core_bounds(tests.IrisTest, AuxCoordTestMixin):
    # Test for AuxCoord.core_bounds() with various types of points and bounds.
    def setUp(self):
        self.setupTestArrays()

    def test_no_bounds(self):
        coord = AuxCoord(self.pts_real)
        result = coord.core_bounds()
        self.assertIsNone(result)

    def test_real_bounds(self):
        coord = AuxCoord(self.pts_real, bounds=self.bds_real)
        result = coord.core_bounds()
        self.assertArraysShareData(
            result,
            self.bds_real,
            "core_bounds() do not share data with the internal array.",
        )

    def test_lazy_bounds(self):
        coord = AuxCoord(self.pts_real, bounds=self.bds_lazy)
        result = coord.core_bounds()
        self.assertEqualLazyArraysAndDtypes(result, self.bds_lazy)

    def test_lazy_bounds_realise(self):
        coord = AuxCoord(self.pts_real, bounds=self.bds_lazy)
        real_bounds = coord.bounds
        result = coord.core_bounds()
        self.assertEqualRealArraysAndDtypes(result, real_bounds)


class Test_lazy_points(tests.IrisTest, AuxCoordTestMixin):
    def setUp(self):
        self.setupTestArrays()

    def test_real_core(self):
        coord = AuxCoord(self.pts_real)
        result = coord.lazy_points()
        self.assertEqualLazyArraysAndDtypes(result, self.pts_lazy)

    def test_lazy_core(self):
        coord = AuxCoord(self.pts_lazy)
        result = coord.lazy_points()
        self.assertIs(result, self.pts_lazy)


class Test_lazy_bounds(tests.IrisTest, AuxCoordTestMixin):
    def setUp(self):
        self.setupTestArrays()

    def test_no_bounds(self):
        coord = AuxCoord(self.pts_real)
        result = coord.lazy_bounds()
        self.assertIsNone(result)

    def test_real_core(self):
        coord = AuxCoord(self.pts_real, bounds=self.bds_real)
        result = coord.lazy_bounds()
        self.assertEqualLazyArraysAndDtypes(result, self.bds_lazy)

    def test_lazy_core(self):
        coord = AuxCoord(self.pts_real, bounds=self.bds_lazy)
        result = coord.lazy_bounds()
        self.assertIs(result, self.bds_lazy)


class Test_has_lazy_points(tests.IrisTest, AuxCoordTestMixin):
    def setUp(self):
        self.setupTestArrays()

    def test_real_core(self):
        coord = AuxCoord(self.pts_real)
        result = coord.has_lazy_points()
        self.assertFalse(result)

    def test_lazy_core(self):
        coord = AuxCoord(self.pts_lazy)
        result = coord.has_lazy_points()
        self.assertTrue(result)

    def test_lazy_core_realise(self):
        coord = AuxCoord(self.pts_lazy)
        coord.points
        result = coord.has_lazy_points()
        self.assertFalse(result)


class Test_has_lazy_bounds(tests.IrisTest, AuxCoordTestMixin):
    def setUp(self):
        self.setupTestArrays()

    def test_real_core(self):
        coord = AuxCoord(self.pts_real, bounds=self.bds_real)
        result = coord.has_lazy_bounds()
        self.assertFalse(result)

    def test_lazy_core(self):
        coord = AuxCoord(self.pts_real, bounds=self.bds_lazy)
        result = coord.has_lazy_bounds()
        self.assertTrue(result)

    def test_lazy_core_realise(self):
        coord = AuxCoord(self.pts_real, bounds=self.bds_lazy)
        coord.bounds
        result = coord.has_lazy_bounds()
        self.assertFalse(result)


class Test_bounds_dtype(tests.IrisTest, AuxCoordTestMixin):
    def test_i16(self):
        test_dtype = np.int16
        coord = AuxCoord([1], bounds=np.array([[0, 4]], dtype=test_dtype))
        result = coord.bounds_dtype
        self.assertEqual(result, test_dtype)

    def test_u16(self):
        test_dtype = np.uint16
        coord = AuxCoord([1], bounds=np.array([[0, 4]], dtype=test_dtype))
        result = coord.bounds_dtype
        self.assertEqual(result, test_dtype)

    def test_f16(self):
        test_dtype = np.float16
        coord = AuxCoord([1], bounds=np.array([[0, 4]], dtype=test_dtype))
        result = coord.bounds_dtype
        self.assertEqual(result, test_dtype)


class Test__getitem__(tests.IrisTest, AuxCoordTestMixin):
    # Test for AuxCoord indexing with various types of points and bounds.
    def setUp(self):
        self.setupTestArrays()

    def test_partial_slice_data_copy(self):
        parent_coord = AuxCoord([1.0, 2.0, 3.0])
        sub_coord = parent_coord[:1]
        values_before_change = sub_coord.points.copy()
        parent_coord.points[:] = -999.9
        self.assertArrayEqual(sub_coord.points, values_before_change)

    def test_full_slice_data_copy(self):
        parent_coord = AuxCoord([1.0, 2.0, 3.0])
        sub_coord = parent_coord[:]
        values_before_change = sub_coord.points.copy()
        parent_coord.points[:] = -999.9
        self.assertArrayEqual(sub_coord.points, values_before_change)

    def test_dtypes(self):
        # Index coords with all combinations of real+lazy points+bounds, and
        # either an int or floating dtype.
        # Check that dtypes remain the same in all cases, taking the dtypes
        # directly from the core points and bounds (as we have no masking).
        for (
            main_coord,
            points_type_name,
            bounds_type_name,
        ) in coords_all_dtypes_and_lazynesses(self, AuxCoord):

            sub_coord = main_coord[:2, 1]

            coord_dtype = main_coord.dtype
            msg = (
                "Indexing main_coord of dtype {} "
                "with {} points and {} bounds "
                "changed dtype of {} to {}."
            )

            sub_points = sub_coord.core_points()
            self.assertEqual(
                sub_points.dtype,
                coord_dtype,
                msg.format(
                    coord_dtype,
                    points_type_name,
                    bounds_type_name,
                    "points",
                    sub_points.dtype,
                ),
            )

            if bounds_type_name != "no":
                sub_bounds = sub_coord.core_bounds()
                main_bounds_dtype = main_coord.bounds_dtype
                self.assertEqual(
                    sub_bounds.dtype,
                    main_bounds_dtype,
                    msg.format(
                        main_bounds_dtype,
                        points_type_name,
                        bounds_type_name,
                        "bounds",
                        sub_bounds.dtype,
                    ),
                )

    def test_lazyness(self):
        # Index coords with all combinations of real+lazy points+bounds, and
        # either an int or floating dtype.
        # Check that lazy data stays lazy and real stays real, in all cases.
        for (
            main_coord,
            points_type_name,
            bounds_type_name,
        ) in coords_all_dtypes_and_lazynesses(self, AuxCoord):

            sub_coord = main_coord[:2, 1]

            msg = (
                "Indexing coord of dtype {} "
                "with {} points and {} bounds "
                "changed laziness of {} from {!r} to {!r}."
            )
            coord_dtype = main_coord.dtype
            sub_points_lazyness = lazyness_string(sub_coord.core_points())
            self.assertEqual(
                sub_points_lazyness,
                points_type_name,
                msg.format(
                    coord_dtype,
                    points_type_name,
                    bounds_type_name,
                    "points",
                    points_type_name,
                    sub_points_lazyness,
                ),
            )

            if bounds_type_name != "no":
                sub_bounds_lazy = lazyness_string(sub_coord.core_bounds())
                self.assertEqual(
                    sub_bounds_lazy,
                    bounds_type_name,
                    msg.format(
                        coord_dtype,
                        points_type_name,
                        bounds_type_name,
                        "bounds",
                        bounds_type_name,
                        sub_bounds_lazy,
                    ),
                )

    def test_real_data_copies(self):
        # Index coords with all combinations of real+lazy points+bounds.
        # In all cases, check that any real arrays are copied by the indexing.
        for (
            main_coord,
            points_lazyness,
            bounds_lazyness,
        ) in coords_all_dtypes_and_lazynesses(self, AuxCoord):

            sub_coord = main_coord[:2, 1]

            msg = (
                "Indexed coord with {} points and {} bounds "
                "does not have its own separate {} array."
            )
            if points_lazyness == "real":
                main_points = main_coord.core_points()
                sub_points = sub_coord.core_points()
                sub_main_points = main_points[:2, 1]
                self.assertEqualRealArraysAndDtypes(
                    sub_points, sub_main_points
                )
                self.assertArraysDoNotShareData(
                    sub_points,
                    sub_main_points,
                    msg.format(points_lazyness, bounds_lazyness, "points"),
                )

            if bounds_lazyness == "real":
                main_bounds = main_coord.core_bounds()
                sub_bounds = sub_coord.core_bounds()
                sub_main_bounds = main_bounds[:2, 1]
                self.assertEqualRealArraysAndDtypes(
                    sub_bounds, sub_main_bounds
                )
                self.assertArraysDoNotShareData(
                    sub_bounds,
                    sub_main_bounds,
                    msg.format(points_lazyness, bounds_lazyness, "bounds"),
                )


class Test_copy(tests.IrisTest, AuxCoordTestMixin):
    # Test for AuxCoord.copy() with various types of points and bounds.
    def setUp(self):
        self.setupTestArrays()

    def test_lazyness(self):
        # Copy coords with all combinations of real+lazy points+bounds, and
        # either an int or floating dtype.
        # Check that lazy data stays lazy and real stays real, in all cases.
        for (
            main_coord,
            points_lazyness,
            bounds_lazyness,
        ) in coords_all_dtypes_and_lazynesses(self, AuxCoord):

            coord_dtype = main_coord.dtype
            copied_coord = main_coord.copy()

            msg = (
                "Copying main_coord of dtype {} "
                "with {} points and {} bounds "
                "changed lazyness of {} from {!r} to {!r}."
            )

            copied_pts_lazyness = lazyness_string(copied_coord.core_points())
            self.assertEqual(
                copied_pts_lazyness,
                points_lazyness,
                msg.format(
                    coord_dtype,
                    points_lazyness,
                    bounds_lazyness,
                    "points",
                    points_lazyness,
                    copied_pts_lazyness,
                ),
            )

            if bounds_lazyness != "no":
                copied_bds_lazy = lazyness_string(copied_coord.core_bounds())
                self.assertEqual(
                    copied_bds_lazy,
                    bounds_lazyness,
                    msg.format(
                        coord_dtype,
                        points_lazyness,
                        bounds_lazyness,
                        "bounds",
                        bounds_lazyness,
                        copied_bds_lazy,
                    ),
                )

    def test_realdata_copies(self):
        # Copy coords with all combinations of real+lazy points+bounds.
        # In all cases, check that any real arrays are copies, not views.
        for (
            main_coord,
            points_lazyness,
            bounds_lazyness,
        ) in coords_all_dtypes_and_lazynesses(self, AuxCoord):

            copied_coord = main_coord.copy()

            msg = (
                "Copied coord with {} points and {} bounds "
                "does not have its own separate {} array."
            )

            if points_lazyness == "real":
                main_points = main_coord.core_points()
                copied_points = copied_coord.core_points()
                self.assertEqualRealArraysAndDtypes(main_points, copied_points)
                self.assertArraysDoNotShareData(
                    main_points,
                    copied_points,
                    msg.format(points_lazyness, bounds_lazyness, "points"),
                )

            if bounds_lazyness == "real":
                main_bounds = main_coord.core_bounds()
                copied_bounds = copied_coord.core_bounds()
                self.assertEqualRealArraysAndDtypes(main_bounds, copied_bounds)
                self.assertArraysDoNotShareData(
                    main_bounds,
                    copied_bounds,
                    msg.format(points_lazyness, bounds_lazyness, "bounds"),
                )


class Test_points__getter(tests.IrisTest, AuxCoordTestMixin):
    def setUp(self):
        self.setupTestArrays()

    def test_mutable_real_points(self):
        # Check that coord.points returns a modifiable array, and changes to it
        # are reflected to the coord.
        data = np.array([1.0, 2.0, 3.0, 4.0])
        coord = AuxCoord(data)
        initial_values = data.copy()
        coord.points[1:2] += 33.1
        result = coord.points
        self.assertFalse(np.all(result == initial_values))

    def test_real_points(self):
        # Getting real points does not change or copy them.
        coord = AuxCoord(self.pts_real)
        result = coord.points
        self.assertArraysShareData(
            result,
            self.pts_real,
            "Points do not share data with the provided array.",
        )

    def test_lazy_points(self):
        # Getting lazy points realises them.
        coord = AuxCoord(self.pts_lazy)
        self.assertTrue(coord.has_lazy_points())
        result = coord.points
        self.assertFalse(coord.has_lazy_points())
        self.assertEqualRealArraysAndDtypes(result, self.pts_real)

    def test_real_points_with_real_bounds(self):
        # Getting real points does not change real bounds.
        coord = AuxCoord(self.pts_real, bounds=self.bds_real)
        coord.points
        result = coord.core_bounds()
        self.assertArraysShareData(
            result,
            self.bds_real,
            "Bounds do not share data with the provided array.",
        )

    def test_real_points_with_lazy_bounds(self):
        # Getting real points does not touch lazy bounds.
        coord = AuxCoord(self.pts_real, bounds=self.bds_lazy)
        coord.points
        self.assertTrue(coord.has_lazy_bounds())

    def test_lazy_points_with_real_bounds(self):
        # Getting lazy points does not affect real bounds.
        coord = AuxCoord(self.pts_lazy, bounds=self.bds_real)
        coord.points
        result = coord.core_bounds()
        self.assertEqualRealArraysAndDtypes(result, self.bds_real)

    def test_lazy_points_with_lazy_bounds(self):
        # Getting lazy points does not touch lazy bounds.
        coord = AuxCoord(self.pts_lazy, bounds=self.bds_lazy)
        coord.points
        self.assertTrue(coord.has_lazy_bounds())


class Test_points__setter(tests.IrisTest, AuxCoordTestMixin):
    def setUp(self):
        self.setupTestArrays()

    def test_real_set_real(self):
        # Setting new real points does not make a copy.
        coord = AuxCoord(self.pts_real)
        new_pts = self.pts_real + 102.3
        coord.points = new_pts
        result = coord.core_points()
        self.assertArraysShareData(
            result,
            new_pts,
            "Points do not share data with the assigned array.",
        )

    def test_fail_bad_shape(self):
        # Setting real points requires matching shape.
        coord = AuxCoord([1.0, 2.0])
        msg = r"Require data with shape \(2,\), got \(3,\)"
        with self.assertRaisesRegex(ValueError, msg):
            coord.points = np.array([1.0, 2.0, 3.0])

    def test_real_set_lazy(self):
        # Setting new lazy points does not make a copy.
        coord = AuxCoord(self.pts_real)
        new_pts = self.pts_lazy + 102.3
        coord.points = new_pts
        result = coord.core_points()
        self.assertEqualLazyArraysAndDtypes(result, new_pts)

    def test_set_points_with_lazy_bounds(self):
        # Setting points does not touch lazy bounds.
        coord = AuxCoord(self.pts_real, bounds=self.bds_lazy)
        new_pts = self.pts_real + 102.3
        coord.points = new_pts
        result = coord.core_bounds()
        self.assertEqualLazyArraysAndDtypes(result, self.bds_lazy)


class Test_bounds__getter(tests.IrisTest, AuxCoordTestMixin):
    def setUp(self):
        self.setupTestArrays()

    def test_mutable_real_bounds(self):
        # Check that coord.bounds returns a modifiable array, and changes to it
        # are reflected to the coord.
        pts_data = np.array([1.5, 2.5])
        bds_data = np.array([[1.4, 1.6], [2.4, 2.6]])
        coord = AuxCoord(pts_data, bounds=bds_data)
        initial_values = bds_data.copy()
        coord.bounds[1:2] += 33.1
        result = coord.bounds
        self.assertFalse(np.all(result == initial_values))

    def test_real_bounds(self):
        # Getting real bounds does not change or copy them.
        coord = AuxCoord(self.pts_real, bounds=self.bds_real)
        result = coord.bounds
        self.assertArraysShareData(
            result,
            self.bds_real,
            "Bounds do not share data with the provided array.",
        )

    def test_lazy_bounds(self):
        # Getting lazy bounds realises them.
        coord = AuxCoord(self.pts_real, bounds=self.bds_lazy)
        self.assertTrue(coord.has_lazy_bounds())
        result = coord.bounds
        self.assertFalse(coord.has_lazy_bounds())
        self.assertEqualRealArraysAndDtypes(result, self.bds_real)

    def test_lazy_bounds_with_lazy_points(self):
        # Getting lazy bounds does not fetch the points.
        coord = AuxCoord(self.pts_lazy, bounds=self.bds_lazy)
        coord.bounds
        self.assertTrue(coord.has_lazy_points())


class Test_bounds__setter(tests.IrisTest, AuxCoordTestMixin):
    def setUp(self):
        self.setupTestArrays()

    def test_set_real_bounds(self):
        # Setting new real bounds does not make a copy.
        coord = AuxCoord(self.pts_real, bounds=self.bds_real)
        new_bounds = self.bds_real + 102.3
        coord.bounds = new_bounds
        result = coord.core_bounds()
        self.assertArraysShareData(
            result,
            new_bounds,
            "Bounds do not share data with the assigned array.",
        )

    def test_fail_bad_shape(self):
        # Setting real points requires matching shape.
        coord = AuxCoord(self.pts_real, bounds=self.bds_real)
        msg = "must be compatible with points shape"
        with self.assertRaisesRegex(ValueError, msg):
            coord.bounds = np.array([1.0, 2.0, 3.0])

    def test_set_lazy_bounds(self):
        # Setting new lazy bounds.
        coord = AuxCoord(self.pts_real, bounds=self.bds_real)
        new_bounds = self.bds_lazy + 102.3
        coord.bounds = new_bounds
        result = coord.core_bounds()
        self.assertEqualLazyArraysAndDtypes(result, new_bounds)

    def test_set_bounds_with_lazy_points(self):
        # Setting bounds does not change lazy points.
        coord = AuxCoord(self.pts_lazy, bounds=self.bds_real)
        new_bounds = self.bds_real + 102.3
        coord.bounds = new_bounds
        self.assertTrue(coord.has_lazy_points())


class Test_convert_units(tests.IrisTest):
    def test_preserves_lazy(self):
        test_bounds = np.array(
            [
                [[11.0, 12.0], [12.0, 13.0], [13.0, 14.0]],
                [[21.0, 22.0], [22.0, 23.0], [23.0, 24.0]],
            ]
        )
        test_points = np.array([[11.1, 12.2, 13.3], [21.4, 22.5, 23.6]])
        lazy_points = as_lazy_data(test_points)
        lazy_bounds = as_lazy_data(test_bounds)
        coord = AuxCoord(points=lazy_points, bounds=lazy_bounds, units="m")
        coord.convert_units("ft")
        self.assertTrue(coord.has_lazy_points())
        self.assertTrue(coord.has_lazy_bounds())
        test_points_ft = Unit("m").convert(test_points, "ft")
        test_bounds_ft = Unit("m").convert(test_bounds, "ft")
        self.assertArrayAllClose(coord.points, test_points_ft)
        self.assertArrayAllClose(coord.bounds, test_bounds_ft)


class TestEquality(tests.IrisTest):
    def test_nanpoints_eq_self(self):
        co1 = AuxCoord([1.0, np.nan, 2.0])
        self.assertEqual(co1, co1)

    def test_nanpoints_eq_copy(self):
        co1 = AuxCoord([1.0, np.nan, 2.0])
        co2 = co1.copy()
        self.assertEqual(co1, co2)

    def test_nanbounds_eq_self(self):
        co1 = AuxCoord([15.0, 25.0], bounds=[[14.0, 16.0], [24.0, np.nan]])
        self.assertEqual(co1, co1)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Unit tests for the :mod:`iris.tests.IrisTest` class."""

# import iris tests first so that some things can be initialised before
# importing anything else
import iris.tests as tests

from abc import ABCMeta, abstractmethod

import numpy as np


class _MaskedArrayEquality(metaclass=ABCMeta):
    def setUp(self):
        self.arr1 = np.ma.array([1, 2, 3, 4], mask=[False, True, True, False])
        self.arr2 = np.ma.array([1, 3, 2, 4], mask=[False, True, True, False])

    @property
    @abstractmethod
    def _func(self):
        pass

    def test_strict_comparison(self):
        # Comparing both mask and data array completely.
        with self.assertRaises(AssertionError):
            self._func(self.arr1, self.arr2, strict=True)

    def test_non_strict_comparison(self):
        # Checking masked array equality and all unmasked array data values.
        self._func(self.arr1, self.arr2, strict=False)

    def test_default_strict_arg_comparison(self):
        self._func(self.arr1, self.arr2)

    def test_nomask(self):
        # Test that an assertion is raised when comparing a masked array
        # containing masked and unmasked values with a masked array with
        # 'nomask'.
        arr1 = np.ma.array([1, 2, 3, 4])
        with self.assertRaises(AssertionError):
            self._func(arr1, self.arr2, strict=False)

    def test_nomask_unmasked(self):
        # Ensure that a masked array with 'nomask' can compare with an entirely
        # unmasked array.
        arr1 = np.ma.array([1, 2, 3, 4])
        arr2 = np.ma.array([1, 2, 3, 4], mask=False)
        self._func(arr1, arr2, strict=False)

    def test_different_mask_strict(self):
        # Differing masks, equal data
        arr2 = self.arr1.copy()
        arr2[0] = np.ma.masked
        with self.assertRaises(AssertionError):
            self._func(self.arr1, arr2, strict=True)

    def test_different_mask_nonstrict(self):
        # Differing masks, equal data
        arr2 = self.arr1.copy()
        arr2[0] = np.ma.masked
        with self.assertRaises(AssertionError):
            self._func(self.arr1, arr2, strict=False)


@tests.iristest_timing_decorator
class Test_assertMaskedArrayEqual(_MaskedArrayEquality, tests.IrisTest_nometa):
    @property
    def _func(self):
        return self.assertMaskedArrayEqual


class Test_assertMaskedArrayEqual__Nonmaasked(tests.IrisTest):
    def test_nonmasked_same(self):
        # Masked test can be used on non-masked arrays.
        arr1 = np.array([1, 2])
        self.assertMaskedArrayEqual(arr1, arr1)

    def test_masked_nonmasked_same(self):
        # Masked test can be used between masked + non-masked arrays, and will
        # consider them equal, when mask=None.
        arr1 = np.ma.masked_array([1, 2])
        arr2 = np.array([1, 2])
        self.assertMaskedArrayEqual(arr1, arr2)

    def test_masked_nonmasked_different(self):
        arr1 = np.ma.masked_array([1, 2])
        arr2 = np.array([1, 3])
        with self.assertRaisesRegex(AssertionError, "Arrays are not equal"):
            self.assertMaskedArrayEqual(arr1, arr2)

    def test_nonmasked_masked_same(self):
        # Masked test can be used between masked + non-masked arrays, and will
        # consider them equal, when mask=None.
        arr1 = np.array([1, 2])
        arr2 = np.ma.masked_array([1, 2])
        self.assertMaskedArrayEqual(arr1, arr2)

    def test_masked_nonmasked_same_falsemask(self):
        # Masked test can be used between masked + non-masked arrays, and will
        # consider them equal, when mask=False.
        arr1 = np.ma.masked_array([1, 2], mask=False)
        arr2 = np.array([1, 2])
        self.assertMaskedArrayEqual(arr1, arr2)

    def test_masked_nonmasked_same_emptymask(self):
        # Masked test can be used between masked + non-masked arrays, and will
        # consider them equal, when mask=zeros.
        arr1 = np.ma.masked_array([1, 2], mask=[False, False])
        arr2 = np.array([1, 2])
        self.assertMaskedArrayEqual(arr1, arr2)


@tests.iristest_timing_decorator
class Test_assertMaskedArrayAlmostEqual(
    _MaskedArrayEquality, tests.IrisTest_nometa
):
    @property
    def _func(self):
        return self.assertMaskedArrayAlmostEqual

    def test_decimal(self):
        arr1, arr2 = np.ma.array([100.0]), np.ma.array([100.003])
        self.assertMaskedArrayAlmostEqual(arr1, arr2, decimal=2)
        with self.assertRaises(AssertionError):
            self.assertMaskedArrayAlmostEqual(arr1, arr2, decimal=3)


if __name__ == "__main__":
    tests.main()
